# pylint: disable=no-self-argument
from typing import Dict, Optional

from pydantic import Field, validator

from suzieq.poller.controller.credential_loader.base_credential_loader import \
    CredentialLoader, CredentialLoaderModel, check_credentials
from suzieq.shared.utils import get_sensitive_data
from suzieq.shared.exceptions import SensitiveLoadError, InventorySourceError


class StaticModel(CredentialLoaderModel):
    """Model for static credential loader
    """
    username: Optional[str]
    password: Optional[str]
    key_passphrase: Optional[str] = Field(alias='key-passphrase')
    keyfile: Optional[str]
    enable_password: Optional[str] = Field(alias='enable-password')

    @validator('password', 'key_passphrase', 'enable_password')
    def validate_sens_field(cls, field):
        """Validate if the sensitive var was passed correctly
        """
        try:
            if field == 'ask':
                # the field is valid, but I cannot ask here the value
                return field
            return get_sensitive_data(field)
        except SensitiveLoadError as e:
            raise ValueError(e)


class StaticLoader(CredentialLoader):
    """Simple credential loader

    Loads the credentials inside all devices
    """

    @classmethod
    def get_data_model(cls):
        return StaticModel

    def init(self, init_data: dict):
        """Initialize parameters common to all devices

        Args:
            init_data (Dict): configuration dictionary

        Raises:
            InventorySourceError: Invalid env argument
        """
        if not self._validate:
            # fix pydantic alias
            init_data['enable_password'] = \
                init_data.pop('enable-password', None)
            init_data['key_passphrase'] = init_data.pop('key-passphrase', None)
        super().init(init_data)

        if self._data.password == 'ask':
            try:
                self._data.password = get_sensitive_data(
                    self._data.password,
                    f'{self.name} Password to login to device: ')
            except SensitiveLoadError as e:
                raise InventorySourceError(f'{self.name} {e}')

        if self._data.enable_password == 'ask':
            try:
                self._data.enable_password = get_sensitive_data(
                    self._data.enable_password,
                    f'{self.name} Insert enable password: '
                )
            except SensitiveLoadError as e:
                raise InventorySourceError(f'{self.name} {e}')

        if self._data.key_passphrase == 'ask':
            try:
                self._data.key_passphrase = get_sensitive_data(
                    self._data.key_passphrase,
                    f'{self.name} Passphrase to decode private key file: '
                )
            except SensitiveLoadError as e:
                raise InventorySourceError(f'{self.name} {e}')

    def load(self, inventory: Dict[str, Dict]):

        for device in inventory.values():
            dev_creds = {
                'ssh_keyfile': device.get('ssh_keyfile') or self._data.keyfile,
                'password': device.get('password') or self._data.password,
                'username': device.get('username') or self._data.username,
                'passphrase': device.get('passphrase')
                or self._data.key_passphrase,
                'enable_password': self._data.enable_password
            }

            self.write_credentials(device, dev_creds)

        check_credentials(inventory)



from typing import List
import numpy as np
import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine


class LldpObj(SqPandasEngine):
    '''Backend class to handle manipulating LLDP table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'lldp'

    def get(self, **kwargs):
        '''A separate get to handle JunOS MACaddr as peer'''

        namespace = kwargs.get('namespace', [])
        columns = kwargs.pop('columns', [])
        use_bond = kwargs.pop('use_bond', "False")
        query_str = kwargs.pop('query_str', '')

        addnl_fields = []
        cols = self.schema.get_display_fields(columns)
        self._add_active_to_fields(kwargs.get('view', self.iobj.view), cols,
                                   addnl_fields)

        if columns == ['default']:
            needed_fields = ['subtype', 'peerMacaddr', 'peerIfindex']
        elif 'peerIfname' in columns:
            needed_fields = ['namespace', 'hostname', 'ifname', 'peerHostname',
                             'subtype', 'peerMacaddr', 'peerIfindex']
        else:
            needed_fields = []

        addnl_fields += [f for f in needed_fields if f not in cols]

        user_query_cols = self._get_user_query_cols(query_str)
        addnl_fields += [x for x in user_query_cols if x not in addnl_fields]

        df = super().get(addnl_fields=addnl_fields, columns=cols, **kwargs)
        if df.empty or (not needed_fields and columns != ['*']):
            return df

        macdf = df.query('subtype.isin(["", "mac address"])')
        if not macdf.empty:
            macs = macdf.peerMacaddr.unique().tolist()
            addrdf = self._get_table_sqobj('address').get(
                namespace=namespace, address=macs,
                columns=['namespace', 'hostname', 'ifname', 'macaddr'])

            if not addrdf.empty:
                df = df.merge(
                    addrdf, how='left',
                    left_on=['namespace', 'peerHostname', 'peerMacaddr'],
                    right_on=['namespace', 'hostname', 'macaddr'],
                    suffixes=['', '_y']) \
                    .fillna({'ifname_y': '-'}) \

                if not df.empty:
                    df['peerIfname'] = np.where(df['peerIfname'] == '-',
                                                df['ifname_y'],
                                                df['peerIfname'])
                    df = df.drop(columns=['hostname_y', 'ifname_y',
                                          'timestamp_y', 'vrf', 'macaddr'],
                                 errors='ignore')

        ifidx_df = df.query('subtype.str.startswith("locally")')
        ifindices = ifidx_df.query('peerIfindex != 0').peerIfindex \
            .unique().tolist()
        if not ifidx_df.empty and ifindices:
            ifdf = self._get_table_sqobj('interfaces').get(
                namespace=namespace, ifindex=ifindices,
                columns=['namespace', 'hostname', 'ifname', 'ifindex'])
            df = df.merge(
                ifdf, how='left',
                left_on=['namespace', 'peerHostname', 'peerIfindex'],
                right_on=['namespace', 'hostname', 'ifindex'],
                suffixes=['', '_y']) \
                .fillna({'ifname_y': '-'})

            if not df.empty:
                df['peerIfname'] = np.where(df['peerIfname'] == '-',
                                            df['ifname_y'], df['peerIfname'])

        if use_bond.lower() == "true":
            df = self._resolve_to_bond(
                df, hostname=kwargs.get('hostname', []))[cols]

        df = self._handle_user_query_str(df, query_str)
        return df.reset_index(drop=True)[cols]

    def summarize(self, **kwargs):
        '''Summarize LLDP info'''
        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('nbrCnt', 'hostname', 'count'),
            ('peerHostnameCnt', 'peerHostname', 'count'),
            ('uniquePeerMgmtIPCnt', 'mgmtIP', 'nunique'),
        ]

        self._summarize_on_add_with_query = [
            ('missingPeerInfoCnt', "mgmtIP == '-' or mgmtIP == ''", 'mgmtIP')
        ]

        return super().summarize(**kwargs)

    def _resolve_to_bond(self, df: pd.DataFrame,
                         hostname=List[str]) -> pd.DataFrame:
        """Change bond/channel member ports into bond name

        With normal port channels or bond interfaces, this routine
        replaces the ifname and peerifname with the appropriate bond
        name.

        Args:
            df: The original LLDP dataframe
            hostname: The hostname selection passed by user

        Returns:
            The LLDP dataframe with the appropriate substitution
        """

        ifdf = self._get_table_sqobj('interfaces').get(
            namespace=df.namespace.unique().tolist(),
            hostname=hostname,
            type=['bond_slave', 'bond'],
            columns=['namespace', 'hostname', 'ifname', 'adminState',
                     'ipAddressList', 'master', 'mtu'])

        if ifdf.empty:
            # Could we do something better, probably. TBD.
            return df

        # OSPF sessions maybe over a port-channel. Make LLDP
        # matchup
        lldp_df = df.merge(ifdf, on=['namespace', 'hostname', 'ifname'],
                           how='left')
        lldp_df = lldp_df.merge(
            ifdf,
            left_on=['namespace', 'peerHostname', 'peerIfname'],
            right_on=['namespace', 'hostname', 'ifname'],
            how='left', suffixes=['', '_peer']) \
            .replace(to_replace={'master': {'bridge': np.nan,
                                            'default': np.nan,
                                            '': np.nan},
                                 'master_peer': {'bridge': np.nan,
                                                 'default': np.nan,
                                                 '': np.nan}})

        lldp_df['ifname'] = np.where(~lldp_df.master.isnull(),
                                     lldp_df.master, lldp_df.ifname)
        lldp_df['peerIfname'] = np.where(~lldp_df.master_peer.isnull(),
                                         lldp_df.master_peer,
                                         lldp_df.peerIfname)
        lldp_df = lldp_df.drop_duplicates(
            subset=['namespace', 'hostname', 'ifname']) \
            .drop(columns=['master', 'master_peer'], errors='ignore')

        return lldp_df



from suzieq.engines.pandas.engineobj import SqPandasEngine


class IfcountersObj(SqPandasEngine):
    '''Backend class to handle ops on interface counters table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'ifCounters'



# pylint: disable=no-name-in-module
from typing import Dict, List, Optional

from pydantic import BaseModel
from suzieq.shared.exceptions import SqPollerConfError
from suzieq.shared.sq_plugin import SqPlugin


def _underscore_to_dash(field_name: str) -> str:
    """Replaces the undescore in the name of the field to a dash
    """
    return field_name.replace('_', '-')


class BasePluginModel(BaseModel):
    """Base model for plugins validation
    """
    name: str

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'
        alias_generator = _underscore_to_dash


class InventoryPluginModel(BasePluginModel):
    """Model for inventory validation
    """
    type: Optional[str]


class ControllerPlugin(SqPlugin):
    """Base class for all controller plugins

    Args:
        SqPlugin ([type]): [description]
    """
    # pylint: disable=unused-argument
    def __init__(self, plugin_conf: Dict, validate: bool = True) -> None:
        self._validate = validate

    @classmethod
    def init_plugins(cls, plugin_conf: Dict, validate: bool = False) \
            -> List[Dict]:
        """Instantiate one or more instances of the current class according
        to the given configuration.
        The validation is set to False by default because when the init_plugins
        function is called, the configuration has already been validated.

        Args:
            plugin_conf (dict): plugin configuration
            validate (bool): validate the plugin during initialization
                             Default: False

        Returns:
            List[Dict]: list of generated plugins
        """

        if plugin_conf is None:
            raise RuntimeError('Plugin configuration cannot be None')

        ptype = plugin_conf.get("type")
        if not ptype:
            raise SqPollerConfError('No default type provided')

        controller_class = cls.get_plugins(plugin_name=ptype)

        if not controller_class:
            raise SqPollerConfError(f"Unknown plugin called {ptype}")

        return [controller_class[ptype](plugin_conf, validate)]

    @classmethod
    def default_type(cls) -> str:
        """Return the default type for a plugin
        """
        raise NotImplementedError

    @classmethod
    def get_data_model(cls) -> BasePluginModel:
        """Return the model used for the data
        """
        raise NotImplementedError



"""
This module contains the logic of the plugin
in charge of creating fake nodes returning
the the data inside an input directory
"""

from typing import Dict, List

from suzieq.poller.worker.inventory.inventory import Inventory
from suzieq.poller.worker.nodes.files import FileNode


class InputDirInventory(Inventory):
    """InputDirInventory is not a real inventory
    source plugin, but it is a debugging instrument,
    since it overrides the Inventory class in order to
    create fake nodes returning the content of the input
    directory
    """

    def __init__(self, add_task_fn, **kwargs) -> None:
        self.input_dir = kwargs.pop('input_dir', None)
        super().__init__(add_task_fn, **kwargs)

    async def build_inventory(self) -> Dict[str, FileNode]:
        """Returns a list containing a single fake node
        returning the data contained in the input directory

        Returns:
            Dict[str, FileNode]: a list containing all the nodes in the
                inventory
        """

        node = FileNode()
        await node.initialize(self.input_dir)
        self._nodes = {node.hostname: node}

        return self._nodes

    async def _get_device_list(self) -> List[Dict]:
        return []



"""
This module contains the logic for instantiaing
and running the OutputWorkers
"""

import asyncio
import logging
from typing import Dict, List

from suzieq.poller.worker.writers.output_worker import OutputWorker
from suzieq.shared.exceptions import SqPollerConfError

logger = logging.getLogger(__name__)


class OutputWorkerManager:
    """OuputWorkerManager is the class in charge of
    taking care of the OutputWorker instantiation and
    that the polling output is persistened by all the
    OutputWorkers.
    """

    def __init__(self,
                 output_types: List[str],
                 output_args: Dict[str, str]) -> None:
        self._output_queue = asyncio.Queue()
        self._output_workers = []

        self._output_types = output_types
        self._output_args = output_args
        self._init_output_workers()

    @property
    def output_queue(self):
        '''Queue between the poller and the writer'''
        return self._output_queue

    @property
    def output_types(self):
        '''Supported output formats'''
        return self._output_types

    @property
    def output_args(self):
        '''Args supplied on init'''
        return self._output_args

    async def run_output_workers(self):
        """Start writing the output data, watching
        the content of the ouput queue and triggering
        a write on all the configured Ouput
        """
        while True:
            try:
                data = await self._output_queue.get()
            except asyncio.CancelledError:
                logger.warning(
                    'OutputWorkerManager: received signal to terminate'
                )
                return

            if not self._output_workers:
                return

            for worker in self._output_workers:
                worker.write_data(data)

    def _init_output_workers(self):
        """Create the appropriate output workers for persisting the
        poller output.
        """
        # Load the available output workers
        worker_types = OutputWorker.get_plugins()
        for otype in self.output_types:
            if otype not in worker_types:
                raise SqPollerConfError(f'{otype} is not a valid output '
                                        f'pick some of {worker_types.keys()}')
            new_worker = worker_types[otype](**self.output_args)
            self._output_workers.append(new_worker)



"""
This module contains all the common logic
of the Suzieq poller output workers
"""
import abc
import logging
import os
from typing import Dict
from suzieq.shared.exceptions import SqPollerConfError

from suzieq.shared.sq_plugin import SqPlugin


class OutputWorker(SqPlugin):
    """OutputWorker is the base class for all the objects
    implementing the logic to persist the output of the polling
    """
    def __init__(self, **kwargs):
        self.type = kwargs.get('type', None)
        self.logger = logging.getLogger(__name__)

        output_dir = kwargs.get('output_dir', None)
        if output_dir:
            self.root_output_dir = output_dir
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)
            elif not os.path.isdir(output_dir):
                raise SqPollerConfError(f'Output directory {output_dir}'
                                        'is not a directory')
        else:
            raise SqPollerConfError('Need mandatory keyword arg: output_dir')

    @abc.abstractmethod
    def write_data(self, data: Dict):
        """Write the data into the ouput source"

        Args:
            data (Dict): dictionary containing the data to store.
        """
        raise NotImplementedError



"""This module manages and coordinates all the plugins

Classes:
    InventoryProvider: manages all the plugins

Functions:
    sq_main(): this function coordinates all the plugins initializing
                and starting each one of them. The plugins are divided
                in types.
"""
import argparse
import asyncio
import logging
import signal
from collections import defaultdict
from pathlib import Path
from typing import Dict, List
from copy import deepcopy

from suzieq.poller.controller.base_controller_plugin import ControllerPlugin
from suzieq.poller.controller.inventory_async_plugin import \
    InventoryAsyncPlugin
from suzieq.poller.worker.services.service_manager import ServiceManager
from suzieq.shared.exceptions import InventorySourceError, SqPollerConfError
from suzieq.shared.utils import sq_get_config_file

logger = logging.getLogger(__name__)

DEFAULT_INVENTORY_PATH = 'suzieq/config/etc/inventory.yaml'


class Controller:
    """This class manages all the plugins set on the configuration files
    """

    def __init__(self, args: argparse.Namespace, config_data: dict) -> None:
        # contains the Plugin objects divided by type
        self._plugin_objects = {}

        # collect basePlugin classes
        self._base_plugin_classes = ControllerPlugin.get_plugins()

        self.sources = []
        self.chunker = None
        self.manager = None

        # Initialize configurations
        self._config = defaultdict(lambda: {})
        self._config.update(config_data.get('poller', {}))

        # Set controller configuration
        # single_run_mode: ['gather', 'process', 'update', 'debug',
        #                   'input-dir']
        #                  tells if the poller should not run forever
        # period: the update timeout of the inventory
        # input-dir: wether to use a directory with some data as input
        # no-coalescer: wether to use the coalescer or not, the coalescer
        #               is always disabled with run-once
        # inventory_timeout: the maximum amount of time to wait for an
        #                    inventory from a source
        self._single_run_mode = args.run_once

        self._input_dir = args.input_dir
        if self._input_dir:
            self._single_run_mode = 'input-dir'

        # If the debug mode is active we need to run the controller only once
        if args.debug:
            self._single_run_mode = 'debug'

        self._no_coalescer = args.no_coalescer
        if self._single_run_mode:
            self._no_coalescer = True

        self._period = args.update_period or \
            self._config.get('update-period', 3600)
        self._inventory_timeout = self._config.get('inventory-timeout', 10)
        self._n_workers = args.workers or self._config.get(
            'manager', {}).get('workers', 1)

        # Validate the arguments
        self._validate_controller_args(args, config_data)

        # Get the inventory
        default_inventory_file = DEFAULT_INVENTORY_PATH
        inventory_file = None

        if not self._input_dir:
            inventory_file = args.inventory or \
                self._config.get('inventory-file') or \
                default_inventory_file
            if not Path(inventory_file).is_file():
                if inventory_file != default_inventory_file:
                    raise SqPollerConfError(
                        f'Inventory file not found at {inventory_file}'
                    )

                raise SqPollerConfError(
                    'Inventory file not found in the default location:'
                    f'{inventory_file}, use -I argument to provide it, '
                    'use -i instead to provide an input directory '
                    'with pre-captured output and simulate an input.'
                )
        else:
            if not Path(self._input_dir).is_dir():
                raise SqPollerConfError(
                    f'{self._input_dir} is not a valid directory'
                )

        source_args = {'single-run-mode': self._single_run_mode,
                       'path': inventory_file}

        manager_args = {'config': sq_get_config_file(args.config),
                        'config-dict': config_data,
                        'debug': args.debug,
                        'input-dir': self._input_dir,
                        'exclude-services': args.exclude_services,
                        'no-coalescer': self._no_coalescer,
                        'output-dir': args.output_dir,
                        'outputs': args.outputs,
                        # `single-run-mode` and `run-once` are different.
                        # The former is an internal variable telling the
                        # poller if it should run and terminate, the other
                        # is a special run mode for the worker.
                        'single-run-mode': self._single_run_mode,
                        'run-once': args.run_once,
                        'service-only': args.service_only,
                        'ssh-config-file': args.ssh_config_file,
                        'workers': self._n_workers
                        }

        # Update configuration with command arguments
        self._config['source'].update(source_args)

        self._config['manager'].update(manager_args)
        if not self._config['manager'].get('type'):
            self._config['manager']['type'] = 'static'

        if not self._config['chunker'].get('type'):
            self._config['chunker']['type'] = 'static'

    @property
    def single_run_mode(self) -> str:
        """Returns the current single-run mode if any, if the poller should
        run forever this function returns None

        Returns:
            [str]: current single-run mode
        """
        return self._single_run_mode

    @property
    def period(self) -> int:
        """Returns the update period of the inventory

        Returns:
            [int]: update period in seconds
        """
        return self._period

    @property
    def inventory_timeout(self) -> int:
        """Returns the maximum amount of time to wait for an inventory source
        retrieving its device list.

        Returns:
            int: inventory timeout in secodns
        """
        return self._inventory_timeout

    def init(self):
        """Loads the provider configuration and the plugins configurations
        and initialize all the plugins
        """

        # Initialize the controller modules
        logger.info('Initializing all the poller controller modules')

        # If the input is a directory sources and chunkers are not needed
        # since there is no need to build and split the inventory
        if not self._input_dir:
            logger.debug('Inizialing sources')

            self.sources = self.init_plugins('source')
            if not self.sources:
                raise SqPollerConfError(
                    "The inventory file doesn't have any source"
                )

            # Initialize chunker module
            logger.debug('Initialize chunker module')

            chunkers = self.init_plugins('chunker')
            if len(chunkers) > 1:
                raise SqPollerConfError(
                    'Only 1 Chunker at a time is supported'
                )
            self.chunker = chunkers[0]

        # initialize pollerManager
        logger.debug('Initialize manager module')

        managers = self.init_plugins('manager')
        if len(managers) > 1:
            raise SqPollerConfError(
                'Only 1 manager at a time is supported'
            )
        self.manager = managers[0]

    def init_plugins(self, plugin_type: str) -> List[ControllerPlugin]:
        """Initialize the controller plugins of type <plugin_type> according
        to the controller configuration

        Args:
            plugin_type (str): type of plugins to initialize

        Raises:
            SqPollerConfError: raised if a wrong configuration is passed

        Returns:
            List[ControllerPlugin]: list of initialized plugins
        """
        plugin_conf = self._config.get(plugin_type) or {}

        base_plugin_class = self._base_plugin_classes.get(plugin_type, None)
        if not base_plugin_class:
            raise SqPollerConfError(f'Unknown plugin type {plugin_type}')

        # Initialize all the instances of the given plugin
        self._plugin_objects[plugin_type] = base_plugin_class.init_plugins(
            plugin_conf)
        return self._plugin_objects[plugin_type]

    async def run(self):
        """Start the device polling phase.

        In the kwargs are passed all the components that must be started

        Args:
            controller (Controller): contains the informations for controller
            and workers
        """
        # When the poller receives a termination signal, we would like
        # to gracefully terminate all the tasks
        loop = asyncio.get_event_loop()
        for s in [signal.SIGTERM, signal.SIGINT]:
            loop.add_signal_handler(
                s, lambda s=s: asyncio.create_task(self._stop()))

        # Start collecting the tasks to launch
        source_tasks = [s.run() for s in self.sources
                        if isinstance(s, InventoryAsyncPlugin)]
        # Check if we need to launch the manager in background
        manager_tasks = []
        if isinstance(self.manager, InventoryAsyncPlugin):
            manager_tasks.append(self.manager.run())

        # Append the synchronization manager
        controller_task = [self._inventory_sync()]

        # Launch all the tasks
        tasks = [asyncio.create_task(t)
                 for t in (source_tasks + manager_tasks + controller_task)]
        try:
            while tasks:
                done, pending = await asyncio.wait(
                    tasks,
                    return_when=asyncio.FIRST_COMPLETED
                )

                tasks = list(pending)
                for task in done:
                    if task.exception():
                        raise task.exception()
                # Ignore completed task if started with single-run mode
                if self._single_run_mode:
                    continue

        except asyncio.CancelledError:
            logger.warning('Received termination signal, terminating...')

    async def _stop(self):
        """Stop the controller"""

        tasks = [t for t in asyncio.all_tasks()
                 if t is not asyncio.current_task()]

        for task in tasks:
            task.cancel()

    async def _inventory_sync(self):
        # With the input directory we do not launch the synchronization loop
        if self._input_dir:
            await self.manager.launch_with_dir()
            return

        while True:
            global_inventory = {}

            for inv_src in self.sources:
                try:
                    cur_inv = deepcopy(await asyncio.wait_for(
                        inv_src.get_inventory(),
                        self._inventory_timeout
                    ))
                except asyncio.TimeoutError:
                    raise InventorySourceError(
                        f'Timeout error: source {inv_src.name} took'
                        'too much time'
                    )

                if cur_inv:
                    cur_inv_count = len(cur_inv)
                    duplicated_devices = [x for x in cur_inv
                                          if x in global_inventory]
                    for dd in duplicated_devices:
                        logger.warning(f'Ignoring duplicated device {dd}')
                        cur_inv.pop(dd)
                    if len(duplicated_devices) == cur_inv_count:
                        logger.warning(
                            f'All {inv_src.name} nodes have been ignored')
                else:
                    logger.warning(
                        f'source {inv_src.name} returned an empty inventory')

                global_inventory.update(cur_inv)

            if not global_inventory:
                raise InventorySourceError('No devices to poll')

            n_pollers = self.manager.get_n_workers(global_inventory)

            inventory_chunks = self.chunker.chunk(global_inventory, n_pollers)

            await self.manager.apply(inventory_chunks)

            if self._single_run_mode:
                break
            await asyncio.sleep(self._period)

    def _validate_controller_args(self, args: argparse.Namespace, cfg: Dict):
        """Validate the arguments given to the Controller instance

        Args:
            args (argparse.Namespace): the arguments given to the controller

        Raises:
            SqPollerConfError: raised if a wrong configuration is passed
        """
        # Check if the timeout and the period are valid
        if self._inventory_timeout < 1:
            raise SqPollerConfError(
                'Invalid inventory timeout: at least 1 second'
            )

        if self._period < 1:
            raise SqPollerConfError(
                'Invalid period: at least one second required'
            )

        if self._n_workers < 0:
            raise SqPollerConfError(
                'At least a worker is required'
            )

        # Check if service list is valid
        ServiceManager.get_service_list(
            args.service_only,
            args.exclude_services,
            cfg['service-directory']
        )



from typing import TypeVar, Dict, Callable, List
from abc import abstractmethod
from collections import defaultdict
import time
from datetime import datetime, timezone
import logging
import random
from http import HTTPStatus
import json
import re
import operator
from urllib.parse import urlparse
import asyncio
from asyncio.subprocess import PIPE, DEVNULL
# pylint: disable=redefined-builtin
from concurrent.futures._base import TimeoutError

from packaging import version as version_parse
import xmltodict
import asyncssh
import aiohttp
from dateparser import parse


from suzieq.poller.worker.services.service import RsltToken
from suzieq.shared.utils import get_timestamp_from_junos_time, known_devtypes
from suzieq.shared.exceptions import SqPollerConfError, UnknownDevtypeError

logger = logging.getLogger(__name__)
IOS_SLEEP_BET_CMDS = 5          # in seconds
TNode = TypeVar('TNode', bound='Node')


# pylint: disable=broad-except, attribute-defined-outside-init
class Node:
    '''Class defining communicating with a device for telemetry

    The device class provides most of the basic functionality necessary to
    communicate with a device. In this base class, we don't know the device
    type i.e. whether the device is a device running EOS, IOSXE etc. Once this
    info is known, the class of the device switches from this base class to
    the device-specific class such as EosNode, IosXENode etc. Every such
    device-specific class has its own specific function for fetching the data
    that'll help determine what commands to issue to a device given its
    version, type etc.

    A device type is either determined automatically (if we use SSH transport)
    or is specified by the user (mandatory for any other transport).

    A device is using this base class then only when it doesn't know its
    device type. So, till the device type is known, we keep invoking the
    method _detect_node_type, backing off every so often. If the device is
    an unsupported device, we do mark it as "unknown" to avoid re-initing
    the info constantly.

    To reduce the overhead of setting up an SSH connection everytime we want
    to communicate, we use persistent SSH sessions to the extent possible. In
    some cases, creating this persistent session is not so easy. This is true
    for IOS devices, and so the device-specific class for those cases handle
    this in their own way. Any transport but SSH does not support persistence.

    Its possible when a device comes back up after a reboot, the device's info
    has changed such as its version. We'll need to re-init this data in order
    to continue fetching data from a device correctly in such a case. This is
    easy to do in case of SSH as the persistent connection is torn down when
    the device reboots. We don't have a good way to handle this with any other
    transport because they don;t have a persistent connection.

    The caller of this class MUST follow the following flow:
        * Create the class
        * Call initialize with the kwargs of at least the IP address of device,
          username, password (or keyfile)
        * Call run to start the node's polling
        * Call post_commands to request node to execute commands on device &
          return data. You'll need to supply a callback fn to call after the
          data is available.
    '''

    # pylint: disable=too-many-statements
    async def initialize(self, **kwargs) -> TNode:
        '''Since the class is largely async, we need a diff init'''

        if not kwargs:
            raise ValueError

        self.devtype = None  # Device type
        self.pvtkey_file = ""     # SSH private keyfile
        self.prev_result = {}  # No updates if nothing changed
        self.nsname = None
        self.svc_cmd_mapping = defaultdict(lambda: {})  # Not used yet
        self.logger = logging.getLogger(__name__)
        self.port = 0
        self.backoff = 15  # secs to backoff
        self.init_again_at = 0  # after this epoch secs, try init again
        self.connect_timeout = kwargs.get('connect_timeout', 15)
        self.cmd_timeout = 10  # default command timeout in seconds
        self.batch_size = 4    # Number of commands to issue in parallel
        self.bootupTimestamp = 0
        self.version = "all"   # OS Version to pick the right defn
        self._service_queue = None
        self._conn = None
        self._tunnel = None
        self._session = None    # Used only by PANOS as of this comment
        self.svcs_proc = set()
        self.error_svcs_proc = set()
        self.ssh_ready = asyncio.Lock()
        self._last_exception = None
        self._exception_timestamp = None
        self._current_exception = None
        self.api_key = None
        self._stdin = self._stdout = self._long_proc = None
        self._retry = True      # set to False if authentication fails

        self.address = kwargs["address"]
        self.hostname = kwargs["address"]  # default till we get hostname
        self.username = kwargs.get("username", "vagrant") or "vagrant"
        self.password = kwargs.get("password", "vagrant") or "vagrant"
        self.transport = kwargs.get("transport", "ssh")
        self.nsname = kwargs.get("namespace", "default")
        self.port = kwargs.get("port", 0)
        self.devtype = None
        self.ssh_config_file = kwargs.get("ssh_config_file", None)
        self.enable_password = kwargs.get('enable_password')

        passphrase = kwargs.get("passphrase", None)
        jump_host = kwargs.get("jump_host", "")
        if jump_host:
            jump_result = urlparse(jump_host)
            self.jump_user = jump_result.username or self.username
            self.jump_host = jump_result.hostname
            self.jump_host_key = None
            if jump_result.port:
                self.jump_port = jump_result.port
            else:
                self.jump_port = 22
            pvtkey_file = kwargs.pop('jump_host_key_file')
            if pvtkey_file:
                self.jump_host_key = self._decrypt_pvtkey(pvtkey_file,
                                                          passphrase)
                if not self.jump_host_key:
                    raise SqPollerConfError('Unable to read private key file'
                                            f' at {pvtkey_file}'
                                            )
        else:
            self.jump_host = None
            self.jump_host_key = None

        self.ignore_known_hosts = kwargs.get('ignore_known_hosts', False)
        self.slow_host = kwargs.get('slow_host', False)
        pvtkey_file = kwargs.get("ssh_keyfile", None)
        if pvtkey_file:
            self.pvtkey = self._decrypt_pvtkey(pvtkey_file, passphrase)
            if not self.pvtkey:
                self.logger.error("ERROR: Falling back to password for "
                                  f"{self.address}:{self.port}")
                self.pvtkey = None
        else:
            self.pvtkey = None

        self._init_service_queue()

        devtype = kwargs.get("devtype", None)
        if devtype:
            self._set_devtype(devtype, '')
            self.logger.warning(
                f'{devtype} supplied for {self.address}:{self.port}')

        if self.transport == "ssh":
            self.port = self.port or 22
            await self._init_ssh(init_dev_data=False)
        elif self.transport == "https":
            self.port = self.port or 443
            if self.devtype:
                # Checking devtype to ensure we didn't get a REST transport
                # without also providing the device type.
                await self._init_rest()

        if not devtype:
            if self.transport != 'ssh':
                self.logger.error(
                    'devtype MUST be specified in inventory file if transport'
                    f'is not ssh for {self.address}')
                return self
            elif self.is_connected:
                # So we have a connection, lets figure out if we know what
                # to do with this device
                await self._detect_node_type()

        # Now we know the dev type, fetch the data we need about the
        # device to get cracking
        if not self.devtype and self._retry:
            self.backoff = min(600, self.backoff * 2) + \
                (random.randint(0, 1000) / 1000)
            self.init_again_at = time.time() + self.backoff
        elif self.devtype not in ['unsupported', None] and self._retry:
            # OK, we know the devtype, now initialize the info we need
            # to start proper operation

            # IOS* closes the connection after the initial cmds
            # are executed. So close the conn at our end too
            # avoiding the initial persistent connection failure that
            # otherwise happens
            if self.devtype in ['iosxe', 'ios', 'iosxr']:
                await self._close_connection()

            await self._fetch_init_dev_data()
            if not self.hostname:
                self.hostname = self.address

        return self

    @property
    def last_exception(self) -> Exception:
        '''Last exception that occurred on this device'''
        return self._last_exception

    @property
    def current_exception(self) -> Exception:
        '''The current exception faced on this device'''
        return self._current_exception

    @current_exception.setter
    def current_exception(self, val: Exception):
        '''current exception setter'''
        self._last_exception = self._current_exception
        self._current_exception = val
        if val:
            self._exception_timestamp = int(time.time()*1000)

    @property
    def is_connected(self):
        '''Is there connectivity to the device at the transport level'''
        return self._conn is not None

    def _decrypt_pvtkey(self, pvtkey_file: str, passphrase: str) -> str:
        """Decrypt private key file"""

        keydata: str = None
        if pvtkey_file:
            try:
                keydata = asyncssh.public_key.read_private_key(pvtkey_file,
                                                               passphrase)
            except Exception as e:  # pylint: disable=broad-except
                self.logger.error(
                    f"ERROR: Unable to read private key file {pvtkey_file}"
                    f"for jump host due to {e}")

        return keydata

    def _init_service_queue(self):
        if not self._service_queue:
            self._service_queue = asyncio.Queue()

    async def _close_connection(self):
        if self.is_connected:
            self._conn.close()
            await self._conn.wait_closed()
        if self._tunnel:
            self._tunnel.close()
            await self._tunnel.wait_closed()

        self._conn = None
        self._tunnel = None
        self._stdin = self._stdout = self._stderr = None

    async def _terminate(self):
        self.logger.warning(
            f'Node: {self.hostname} received signal to terminate')
        await self._close_connection()
        return

    def _create_error(self, cmd) -> dict:
        data = {'error': str(self.current_exception)}
        if isinstance(self.current_exception, TimeoutError):
            status = HTTPStatus.REQUEST_TIMEOUT
        elif isinstance(self.current_exception, asyncssh.misc.ProtocolError):
            status = HTTPStatus.FORBIDDEN
        elif isinstance(self.current_exception,
                        asyncssh.misc.PermissionDenied):
            status = HTTPStatus.FORBIDDEN
        elif hasattr(self.current_exception, 'code'):
            status = self.current_exception.code
        else:
            status = -1
        return self._create_result(cmd, status, data)

    def _create_result(self, cmd, status, data) -> dict:
        if self.port in [22, 443]:
            # Ignore port if defaults (SSH or HTTPS)
            addrstr = self.address
        else:
            addrstr = f'{self.address}:{self.port}'
        result = {
            "status": status,
            "timestamp": int(datetime.now(tz=timezone.utc).timestamp() * 1000),
            "cmd": cmd,
            "devtype": self.devtype,
            "namespace": self.nsname,
            "hostname": self.hostname,
            "address": addrstr,
            "version": self.version,
            "data": data,
        }
        return result

    async def _parse_device_type_hostname(self, output, _) -> None:
        devtype = ""
        hostname = None

        if output[0]["status"] == 0:
            # don't keep trying if we're connected to an unsupported dev
            devtype = 'unsupported'
            data = output[0]["data"]
            version_str = data

            if 'Arista' in data or 'vEOS' in data:
                devtype = "eos"
            elif "JUNOS " in data:
                model = re.search(r'Model:\s+(\S+)', data)
                if model:
                    if model.group(1).startswith(('mx', 'vmx')):
                        devtype = 'junos-mx'
                    elif 'qfx10' in model.group(1):
                        devtype = 'junos-qfx10k'
                    elif 'qfx' in model.group(1):
                        devtype = 'junos-qfx'
                    elif 'ex' in model.group(1):
                        devtype = 'junos-ex'
                    elif model.group(1).startswith(('srx', 'vSRX')):
                        devtype = 'junos-es'
                if not devtype:
                    devtype = "junos"
            elif "NX-OS" in data:
                devtype = "nxos"
            elif "SONiC" in data:
                devtype = "sonic"
            elif "Cisco IOS XR" in data:
                devtype = "iosxr"
            elif any(x in data for x in ["Cisco IOS XE", "Cisco IOS-XE"]):
                devtype = "iosxe"
            elif "Cisco IOS Software" in data:
                devtype = "ios"
            else:
                self.logger.info(
                    f'{self.address}: Got unrecognized device show version: '
                    f'{data}')

            if devtype.startswith("junos"):
                hmatch = re.search(r'Hostname:\s+(\S+)\n', data)
                if hmatch:
                    hostname = hmatch.group(1)
            elif devtype == "nxos":
                hgrp = re.search(r'[Dd]evice\s+name:\s+(\S+)', data)
                if hgrp:
                    hostname = hgrp.group(1)
                else:
                    hostname = self.address
            elif devtype == "eos":
                # We'll fill in the hostname when the node gets re-init
                hostname = None
            elif devtype in ["iosxe", "ios"]:
                matchval = re.search(r'(\S+)\s+uptime', output[0]['data'])
                if matchval:
                    hostname = matchval.group(1).strip()
                else:
                    hostname = self.address
            elif devtype != "iosxr":
                hgrp = re.search(r'(\S+)\s+uptime\s+is', data)
                if hgrp:
                    hostname = hgrp.group(1)
                else:
                    hostname = self.address

        elif (len(output) > 1) and (output[1]["status"] == 0):
            devtype = 'unsupported'
            data = output[1]["data"]
            if data:
                if "Cumulus Linux" in data:
                    devtype = "cumulus"
                else:
                    devtype = "linux"

                version_str = data
                # Hostname is the last line of the output
                if len(data.strip()) > 0:
                    hostname = data.splitlines()[-1].strip()

        if devtype == 'unsupported':
            if not self.current_exception:
                self.logger.info(
                    f'Unable to determine devtype for '
                    f'{self.address}:{self.port}')
                self._set_devtype(devtype, version_str)
                self._set_hostname(self.address)
                await self._close_connection()
                self.current_exception = UnknownDevtypeError()
        elif devtype:
            self.logger.warning(
                f'Detected {devtype} for {self.address}:{self.port},'
                f' {hostname}')
            self._set_devtype(devtype, version_str)
            self._set_hostname(hostname)
            self.current_exception = None

    async def _detect_node_type(self):
        '''Figure out what type of device this is: EOS, NXOS etc.

        Its only available if we're using ssh as transport
        '''

        devtype = None
        hostname = None

        if self.transport in ["ssh", "local"]:
            try:
                await self._get_device_type_hostname()
                # The above fn calls results in invoking a callback fn
                # that sets the device type
                devtype = self.devtype
            except Exception:
                devtype = None

            if not devtype:
                self.logger.debug(
                    f"no devtype for {self.hostname} {self.current_exception}")
                return
            else:
                self._set_devtype(devtype, '')
                self._set_hostname(hostname)
        else:
            self.logger.error(
                f'Non-SSH transport node {self.address}:{self.port} '
                'has no devtype specified. Node will not be polled')

    async def _get_device_type_hostname(self):
        """Determine the type of device we are talking to if using ssh/local"""
        # There isn't that much of a difference in running two commands versus
        # running them one after the other as this involves an additional ssh
        # setup time. show version works on most networking boxes and
        # hostnamectl on Linux systems. That's all we support today.
        await self._exec_cmd(self._parse_device_type_hostname,
                             ["show version",
                              "cat /etc/os-release && hostname"],
                             None, 'text', only_one=True)

    def _set_devtype(self, devtype: str, version_str: str) -> None:
        """Change the class based on the device type"""

        if not devtype:
            return

        if devtype == 'unsupported':
            self.devtype = devtype
            return

        if devtype not in known_devtypes():
            self.logger.error(f'An unknown devtype {devtype} is being added.'
                              f' This will cause problems. '
                              f'Node {self.address}:{self.port}')
            raise ValueError

        if self.devtype != devtype:
            self.devtype = devtype
            if self.devtype == "cumulus":
                self.__class__ = CumulusNode
            elif self.devtype == "eos":
                self.__class__ = EosNode
            elif self.devtype == "iosxe":
                self.__class__ = IosXENode
            elif self.devtype == "iosxr":
                self.__class__ = IosXRNode
            elif self.devtype == "ios":
                self.__class__ = IOSNode
            elif self.devtype.startswith("junos"):
                self.__class__ = JunosNode
            elif self.devtype == "nxos":
                self.__class__ = NxosNode
            elif self.devtype.startswith("sonic"):
                self.__class__ = SonicNode
            elif self.devtype == "panos":
                self.__class__ = PanosNode
            elif self.devtype == "linux":
                self.__class__ = LinuxNode

        # Now invoke the class specific NOS version extraction
        if version_str:
            self._extract_nos_version(version_str)

    def _set_hostname(self, hostname=None):
        '''Set node hostname'''
        if hostname:
            self.hostname = hostname

    async def _init_jump_host_connection(
            self,
            options: asyncssh.SSHClientConnectionOptions) -> None:
        """Initialize jump host connection if necessary

        Args:
            options (asyncssh.SSHClientConnectionOptions): non-jump host opt
        """

        if self._tunnel:
            return

        if self.jump_host_key:
            jump_host_options = asyncssh.SSHClientConnectionOptions(
                client_keys=self.jump_host_key,
                connect_timeout=self.connect_timeout,
            )

            if self.ignore_known_hosts:
                jump_host_options = asyncssh.SSHClientConnectionOptions(
                    options=jump_host_options,
                    known_hosts=None
                )
            if self.ssh_config_file:
                jump_host_options = asyncssh.SSHClientConnectionOptions(
                    options=jump_host_options,
                    config_file=[self.ssh_config_file]
                )
        else:
            jump_host_options = options

        try:
            if self.jump_host:
                self.logger.info(
                    'Using jump host: %s, with username: %s, and port: %s',
                    self.jump_host, self.jump_user, self.jump_port
                )
                self._tunnel = await asyncssh.connect(
                    self.jump_host, port=self.jump_port,
                    options=jump_host_options, username=self.jump_user)
                self.logger.info(
                    'Connection to jump host %s succeeded', self.jump_host)

        except Exception as e:  # pylint: disable=broad-except
            self.logger.error(
                f'Cannot connect to jump host: {self.jump_host} ({e})')
            self.current_exception = e
            self._conn = None
            self._tunnel = None

    def _init_ssh_options(self) -> asyncssh.SSHClientConnectionOptions:
        """Build out the asycnssh options as specified by user config

        This is a routine because its used in multiple places.
        Returns:
            asyncssh.SSHClientConnectionOptions: [description]
        """
        options = asyncssh.SSHClientConnectionOptions(
            connect_timeout=self.connect_timeout,
            username=self.username,
            agent_identities=self.pvtkey if self.pvtkey else None,
            client_keys=self.pvtkey if self.pvtkey else None,
            password=self.password if not self.pvtkey else None,
            kex_algs='+diffie-hellman-group1-sha1',  # for older boxes
            encryption_algs='+aes256-cbc',           # for older boxes
        )
        if self.ignore_known_hosts:
            options = asyncssh.SSHClientConnectionOptions(
                options=options,
                known_hosts=None,
            )
        if self.ssh_config_file:
            options = asyncssh.SSHClientConnectionOptions(
                options=options,
                config=[self.ssh_config_file],
            )

        return options

    async def _init_ssh(self, init_dev_data=True, use_lock=True) -> None:
        '''Setup a persistent SSH connection to our node.

        In some cases such as IOSXE/XR, this may not actually setup a
        persistent SSH session, but does the basics of setting one up.

        use_lock is a critical parameter to avoid deadlocks. If too
        many services attempt to init_ssh at the same time, we'll cause
        much heartache. So, we acquire a lock to ensure this is done only
        once.

        However, when the calling party has additional work to do, the
        caller MUST acquire the lock and pass use_lock=False. And the
        caller MUST ensure they release the lock.
        '''
        if self._retry and not self._conn:
            if use_lock:
                await self.ssh_ready.acquire()

            # Someone else may have already succeeded in getting the SSH conn
            if self.is_connected or not self._retry:
                if use_lock:
                    self.ssh_ready.release()
                return

            options = self._init_ssh_options()
            if self.jump_host and not self._tunnel:
                await self._init_jump_host_connection(options)
                if not self._tunnel:
                    if use_lock:
                        self.ssh_ready.release()
                    return

            try:
                if self._tunnel:
                    self._conn = await self._tunnel.connect_ssh(
                        self.address, port=self.port,
                        username=self.username,
                        options=options)
                else:
                    self._conn = await asyncssh.connect(
                        self.address,
                        username=self.username,
                        port=self.port,
                        options=options)
                self.logger.info(
                    f"Connected to {self.address}:{self.port} at "
                    f"{time.time()}")
                if init_dev_data:
                    await self._fetch_init_dev_data()
            except Exception as e:  # pylint: disable=broad-except
                if isinstance(e, asyncssh.HostKeyNotVerifiable):
                    self.logger.error(
                        f'Unable to connect to {self.address}: {self.port}, '
                        'host key is unknown. If you do not need to verify '
                        'the host identity, add "ignore-known-hosts: True" in '
                        'the device section of the inventory')
                elif isinstance(e, asyncssh.misc.PermissionDenied):
                    self.logger.error(
                        f'Authentication failed to {self.address}. '
                        'Not retrying to avoid locking out user. Please '
                        'restart poller with proper authentication')
                    self._retry = False
                else:
                    self.logger.error('Unable to connect to '
                                      f'{self.address}:{self.port}, {e}')
                self.current_exception = e
                await self._close_connection()
                self._conn = None
            finally:
                if use_lock:
                    self.ssh_ready.release()

    @abstractmethod
    async def _init_rest(self):
        '''Check that connectivity exists and works'''
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    # pylint: disable=unused-argument
    async def _local_gather(self, service_callback: Callable,
                            cmd_list: List[str], cb_token: RsltToken,
                            oformat: str, timeout: int):
        """Use local command execution to execute the commands requested

        This function takes the raw list of commands to execute on the device
        locally and calls the callback function with the data returned.

        Args:
            service_callback: The callback fn to call when we have the data
                for the commands requested (or error for any given command)
            cmd_list: The list of commands to be executed on the remote device
            cb_token: The callback token to be returned to the callback fn
            oformat: The output format we're expecting the data in: text, json
                or mixed
            timeout: How long to wait for the commands to complete, in secs
            only_one: Run till the first command in the list succeeds and
                then return
        """
        result = []
        for cmd in cmd_list:
            proc = await asyncio.create_subprocess_shell(cmd, stdout=PIPE,
                                                         stderr=PIPE)

            try:
                stdout, stderr = await asyncio.wait_for(
                    proc.communicate(), timeout=self.cmd_timeout)

                if not proc.returncode:
                    d = stdout.decode('ascii', 'ignore')
                    result.append(self._create_result(
                        cmd, proc.returncode, d))
                else:
                    d = stderr('ascii', 'ignore')
                    result.append(self._create_error(cmd))

            except asyncio.TimeoutError as e:
                self.current_exception = e
                result.append(self._create_error(cmd))

        await service_callback(result, cb_token)

    # pylint: disable=unused-argument
    async def _ssh_gather(self, service_callback: Callable,
                          cmd_list: List[str], cb_token: RsltToken,
                          oformat: str, timeout: int, only_one: bool = False):
        """Use SSH to execute the commands requested

        This function takes the raw list of commands to execute on the device
        via SSH and calls the callback function with the data returned.

        Args:
            service_callback: The callback fn to call when we have the data
                for the commands requested (or error for any given command)
            cmd_list: The list of commands to be executed on the remote device
            cb_token: The callback token to be returned to the callback fn
            oformat: The output format we're expecting the data in: text, json
                or mixed
            timeout: How long to wait for the commands to complete, in secs
            only_one: Run till the first command in the list succeeds and
                then return
        """
        result = []

        if cmd_list is None:
            await service_callback({}, cb_token)

        if not self._conn:
            await self._init_ssh()
            if not self._conn:
                for cmd in cmd_list:
                    self.logger.error(
                        "Unable to connect to node %s cmd %s",
                        self.hostname, cmd)
                    result.append(self._create_error(cmd))
                await service_callback(result, cb_token)
                return

        if isinstance(cb_token, RsltToken):
            cb_token.node_token = self.bootupTimestamp

        timeout = timeout or self.cmd_timeout
        for cmd in cmd_list:
            try:
                output = await asyncio.wait_for(self._conn.run(cmd),
                                                timeout=timeout)
                if self.current_exception:
                    self.logger.info(
                        '%s recovered from previous exception', self.hostname)
                    self.current_exception = None
                result.append(self._create_result(
                    cmd, output.exit_status, output.stdout))
                if (output.exit_status == 0) and only_one:
                    break
            except Exception as e:
                self.current_exception = e
                result.append(self._create_error(cmd))
                if not isinstance(e, asyncio.TimeoutError):
                    self.logger.error(
                        "%s output for %s failed due to %s", cmd,
                        self.hostname, e)
                    await self._close_connection()
                else:
                    self.logger.error(
                        "%s output for %s failed due to timeout", cmd,
                        self.hostname)

                break

        await service_callback(result, cb_token)

    async def _exec_cmd(self, service_callback, cmd_list, cb_token,
                        oformat='json', timeout=None, only_one=False,
                        reconnect=True):
        '''Routine to execute a given set of commands on device

        if only_one is True, commands are executed until the first one that
        succeeds, and the rest are ignored.
        '''

        if self.transport == "ssh":
            await self._ssh_gather(service_callback, cmd_list, cb_token,
                                   oformat, timeout, only_one)
        elif self.transport == "https":
            await self._rest_gather(service_callback, cmd_list,
                                    cb_token, oformat, timeout)
        elif self.transport == "local":
            await self._local_gather(service_callback, cmd_list,
                                     cb_token, oformat, timeout)
        else:
            self.logger.error(
                "Unsupported transport %s for node %s",
                self.transport, self.hostname)
        return

    # pylint: disable=too-many-nested-blocks
    async def _exec_service(self, service_callback, svc_defn: dict,
                            cb_token: RsltToken):
        '''Routine that determines cmdlist to be executed for given service

        And invokes appropriate transport routine to execute the cmds and
        return the data to the service that requested it.
        '''
        result = []  # same type as gather function
        cmd = None
        if not svc_defn:
            return result

        if not self.devtype and self._retry:
            if self.init_again_at < time.time():
                await self._detect_node_type()

        if not self.devtype:
            result.append(self._create_error(svc_defn.get("service", "-")))
            return await service_callback(result, cb_token)

        if self.devtype == 'unsupported':
            # Service code 418 means I'm a teapot in http status codes
            # in other words, I won't brew coffee because I'm a teapot
            result.append(self._create_result(
                svc_defn, 418, "No service definition"))
            self.error_svcs_proc.add(svc_defn.get("service"))
            return await service_callback(result, cb_token)

        # Update our boot time value into the callback token
        if cb_token:
            cb_token.bootupTimestamp = self.bootupTimestamp

        self.svcs_proc.add(svc_defn.get("service"))
        use = svc_defn.get(self.hostname, None)
        if not use:
            use = svc_defn.get(self.devtype, {})
        if not use:
            if svc_defn.get("service") not in self.error_svcs_proc:
                result.append(self._create_result(
                    svc_defn, HTTPStatus.NOT_FOUND, "No service definition"))
                self.error_svcs_proc.add(svc_defn.get("service"))
            return await service_callback(result, cb_token)

        # TODO This kind of logic should be encoded in config and node
        # shouldn't have to know about it
        if "copy" in use:
            use = svc_defn.get(use.get("copy"))

        if use:
            if isinstance(use, list):
                # There's more than one version here, we have to pick ours
                for item in use:
                    if item['version'] != "all":
                        os_version = item['version']
                        opdict = {'>': operator.gt, '<': operator.lt,
                                  '>=': operator.ge, '<=': operator.le,
                                  '=': operator.eq, '!=': operator.ne}
                        op = operator.eq

                        for elem, val in opdict.items():
                            if os_version.startswith(elem):
                                os_version = os_version.replace(
                                    elem, '').strip()
                                op = val
                                break

                        if op(version_parse.LegacyVersion(self.version),
                                version_parse.LegacyVersion(os_version)):
                            cmd = item.get('command', None)
                            use = item
                            break
                    else:
                        cmd = item.get("command", None)
                        use = item
                        break
            else:
                cmd = use.get("command", None)

        if not cmd:
            return result

        oformat = use.get('format', 'json')
        if not isinstance(cmd, list):
            if use.get('textfsm'):
                oformat = 'text'
            cmdlist = [cmd]
        else:
            # TODO: Handling format for the multiple cmd case
            cmdlist = [x.get('command', '') for x in cmd]

        await self._exec_cmd(service_callback, cmdlist, cb_token,
                             oformat=oformat, timeout=cb_token.timeout)

    @abstractmethod
    async def _fetch_init_dev_data(self):
        """Start data fetch to initialize the class with specific device attrs

        This function initiates the process of fetching critical pieces
        of info about the device such as version, os type and hostname.
        The version string is specifically used to identify which command
        needs to be executed on a device.

        This is where the list of commands specific to the device for
        extracting the said info is specified.
        """
        raise NotImplementedError(
            f'{self.address}: initing base Node class')

    @abstractmethod
    async def _parse_init_dev_data(self, output: List,
                                   cb_token: RsltToken) -> None:
        """Parse the version, uptime and hostname info from the output

        This function is the callback that extracts the uptime and hostname
        from the data fetched from the device. This function calls the device
        specific version extraction function to extract the version.

        Args:
            output: The list of outputs, one per command specified
            cb_token: The callback token we passed to the data fetcher function
        """
        raise NotImplementedError(
            f'{self.address}: parsing init base Node class')

    @abstractmethod
    def _extract_nos_version(self, data: str) -> str:
        """Extract the version string from the output passed

        Args:
            data: The output of the command to search for the version
                string

        Returns:
            The version as a string
        """
        raise NotImplementedError(
            f'{self.address}: extracting NOS in init base Node class')

    @abstractmethod
    async def _rest_gather(self, service_callback: Callable,
                           cmd_list: List[str], cb_token: RsltToken,
                           oformat: str = "json", timeout: int = None):
        """Use HTTP(s) to execute the commands requested

        This function takes the raw list of commands to execute on the device
        via a REST API and calls the callback function with the data returned.

        Args:
            service_callback: The callback fn to call when we have the data
                for the commands requested (or error for any given command)
            cmd_list: The list of commands to be executed on the remote device
            cb_token: The callback token to be returned to the callback fn
            oformat: The output format we're expecting the data in: text, json
                or mixed
            timeout: How long to wait for the commands to complete, in secs
            only_one: Run till the first command in the list succeeds and
                then return
        """
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    def post_commands(self, service_callback: Callable, svc_defn: Dict,
                      cb_token: RsltToken):
        """Post commands to the device for servicing

        Whenever any user of a device wishes to issue a command to
        this device, they call this function with a list of commands
        to be executed, and a callback function that is invoked when
        the commands have been executed and either there's an error or
        data is returned. Multiple commands imply the result contains
        a list of responses, one for each command that was passed. The
        result itself is a dictionary which includes the command for
        which this is the result. If the user wishes to have a token
        returned to the callback function, they can pass that via the
        cb_token parameter.

        Args:
            service_callback: The callback function
            svc_defn: The service definition file from which the
                appropriate command will be extracted for the device
            cb_token: The callback token to be passed to the callback
                function
        """
        if cb_token:
            cb_token.nodeQsize = self._service_queue.qsize()
        self._service_queue.put_nowait([service_callback, svc_defn, cb_token])

    async def run(self):
        '''Main workhorse routine for Node

        Waits for services to be executed on node via service queue,
        and executes them.
        '''

        tasks = []
        while True:
            while len(tasks) < self.batch_size:
                try:
                    request = await self._service_queue.get()
                except asyncio.CancelledError:
                    await self._terminate()
                    return

                if request:
                    tasks.append(self._exec_service(
                        request[0], request[1], request[2]))
                    self.logger.debug(
                        f"Scheduling {request[2].service} for execution")
                if self._service_queue.empty():
                    break

            if tasks:
                _, pending = await asyncio.wait(
                    tasks, return_when=asyncio.FIRST_COMPLETED)

                tasks = list(pending)


class EosNode(Node):
    '''EOS Node specific implementation'''

    async def _init_rest(self):
        '''Check that connectivity and authentication works'''

        timeout = self.cmd_timeout

        auth = aiohttp.BasicAuth(self.username,
                                 password=self.password or 'vagrant')
        url = f"https://{self.address}:{self.port}/command-api"

        try:
            async with aiohttp.ClientSession(
                    auth=auth, timeout=self.connect_timeout,
                    connector=aiohttp.TCPConnector(ssl=False)) as session:
                async with session.post(url, timeout=timeout) as response:
                    _ = response.status
        except Exception as e:
            self.logger.error(
                f'Unable to connect to {self.address}:{self.port}, '
                f'error: {str(e)}')

    async def _fetch_init_dev_data(self):

        if self.transport == 'https':
            cmdlist = ["show version", "show hostname"]
        else:
            cmdlist = ["show version|json", "show hostname|json"]
        await self._exec_cmd(self._parse_init_dev_data, cmdlist,
                             None, reconnect=False)

    async def _ssh_gather(self, service_callback, cmd_list, cb_token, oformat,
                          timeout, only_one=False):
        """Use SSH to execute the commands requested

        This function takes the raw list of commands to execute on the device
        via SSH and calls the callback function with the data returned.

        Args:
            service_callback: The callback fn to call when we have the data
                for the commands requested (or error for any given command)
            cmd_list: The list of commands to be executed on the remote device
            cb_token: The callback token to be returned to the callback fn
            oformat: The output format we're expecting the data in: text, json
                or mixed
            timeout: How long to wait for the commands to complete, in secs
            only_one: Run till the first command in the list succeeds and
                then return
        """
        # We need to add the JSON option for all commands that support JSON
        # output since the command provided assumes REST API
        newcmd_list = []
        for cmd in cmd_list:
            if (oformat == "json") and not cmd.endswith('json'):
                cmd += '| json'

            newcmd_list.append(cmd)

        return await super()._ssh_gather(service_callback, newcmd_list,
                                         cb_token, oformat, timeout, only_one)

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None):

        result = []
        if not cmd_list:
            return result

        timeout = timeout or self.cmd_timeout

        now = int(datetime.now(tz=timezone.utc).timestamp() * 1000)
        auth = aiohttp.BasicAuth(self.username,
                                 password=self.password or 'vagrant')
        data = {
            "jsonrpc": "2.0",
            "method": "runCmds",
            "id": int(now),
            "params": {"version": 1, "format": oformat, "cmds": cmd_list},
        }
        headers = {"Content-Type": "application/json"}
        if self.port:
            url = "https://{}:{}/command-api".format(self.address, self.port)
        else:
            url = "https://{}:{}/command-api".format(self.address, self.port)

        output = []
        status = 200  # status OK

        try:
            async with aiohttp.ClientSession(
                    auth=auth, conn_timeout=self.connect_timeout,
                    read_timeout=timeout,
                    connector=aiohttp.TCPConnector(ssl=False)) as session:
                async with session.post(url, json=data,
                                        timeout=timeout,
                                        headers=headers) as response:
                    status = response.status
                    if status == HTTPStatus.OK:
                        json_out = await response.json()
                        if "result" in json_out:
                            output.extend(json_out["result"])
                        else:
                            output.extend(
                                json_out["error"].get('data', []))

                        for i, cmd in enumerate(cmd_list):
                            result.append(
                                {
                                    "status": status,
                                    "timestamp": now,
                                    "cmd": cmd,
                                    "devtype": self.devtype,
                                    "namespace": self.nsname,
                                    "hostname": self.hostname,
                                    "address": self.address,
                                    "data":
                                    output[i]
                                    if isinstance(output, list) else output,
                                }
                            )
                    else:
                        for cmd in cmd_list:
                            result.append(self._create_error(cmd))
                        self.logger.error(
                            '(REST), Communication with %s:%s failed'
                            ' due to %s', self.address, self.port,
                            response.status)
        except Exception as e:
            self.current_exception = e
            for cmd in cmd_list:
                result.append(self._create_error(cmd))
            self.logger.error("ERROR: (REST) Unable to communicate with node "
                              "%s:%d due to %s", self.address, self.port,
                              e)

        await service_callback(result, cb_token)

    async def _parse_init_dev_data(self, output, cb_token) -> None:

        if output[0]["status"] == 0 or output[0]["status"] == 200:
            if self.transport == 'ssh':
                try:
                    data = json.loads(output[0]["data"])
                except json.JSONDecodeError:
                    self.logger.error(
                        f'nodeinit: Error decoding JSON for '
                        f'{self.address}:{self.port}')
                    return
            else:
                data = output[0]["data"]
            self.bootupTimestamp = data["bootupTimestamp"]
            self._extract_nos_version(data)
            if not self.version:
                self.logger.error(f'nodeinit: Error getting version for '
                                  f'{self.address}: {self.port}')

        if (len(output) > 1) and (output[1]["status"] in [0, 200]):
            if self.transport == 'ssh':
                try:
                    data = json.loads(output[1]["data"])
                except json.JSONDecodeError:
                    self.logger.error(
                        f'nodeinit: Error decoding JSON for '
                        f'{self.address}:{self.port}')
                    return
            else:
                data = output[1]["data"]
            self.hostname = data["fqdn"]

    def _extract_nos_version(self, data) -> None:
        # < 4.27 or so, the cases were different.
        if isinstance(data, str):
            match = re.search(r'Software Image Version:\s+(\S+)', data,
                              re.IGNORECASE)
            if match:
                self.version = match.group(1).strip()
            else:
                self.logger.warning(
                    f'Cannot parse version from {self.address}:{self.port}')
                self.version = "all"
        else:
            self.version = data['version']


class CumulusNode(Node):
    '''Cumulus Node specific implementation'''

    async def _fetch_init_dev_data(self):
        """Fill in the boot time of the node by executing certain cmds"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["cat /proc/uptime", "hostname",
                              "cat /etc/os-release"], None, 'text')

    async def _parse_init_dev_data(self, output, _) -> None:
        """Parse the uptime command output"""

        if output[0]["status"] == 0:
            upsecs = output[0]["data"].split()[0]
            self.bootupTimestamp = int(int(time.time()*1000)
                                       - float(upsecs)*1000)
        if (len(output) > 1) and (output[1]["status"] == 0):
            data = output[1].get("data", '')
            hostname = data.splitlines()[0].strip()
            self.hostname = hostname

        if (len(output) > 2) and (output[2]["status"] == 0):
            data = output[2].get("data", '')
            self._extract_nos_version(data)

    def _extract_nos_version(self, data: str) -> None:
        """Extract the version from the output of /etc/os-release

        Args:
            data (str): output of show version

        Returns:
            str: version
        """
        version_str = re.search(r'VERSION_ID=(\S+)', data)
        if version_str:
            self.version = version_str.group(1).strip()
        else:
            self.version = "all"
            self.logger.error(
                f'Cannot parse version from {self.address}:{self.port}')

    async def _init_rest(self):
        '''Check that connectivity exists and works'''

        auth = aiohttp.BasicAuth(self.username, password=self.password)
        url = "https://{0}:{1}/nclu/v1/rpc".format(self.address, self.port)
        headers = {"Content-Type": "application/json"}

        try:
            async with aiohttp.ClientSession(
                    auth=auth, timeout=self.cmd_timeout,
                    connector=aiohttp.TCPConnector(ssl=False),
            ) as session:
                async with session.post(url, headers=headers) as response:
                    _ = response.status
        except Exception as e:
            self.current_exception = e
            self.logger.error("ERROR: (REST) Unable to communicate with node "
                              "%s:%d due to %s", self.address, self.port,
                              e)

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat='json', timeout=None):

        result = []
        if not cmd_list:
            return result

        auth = aiohttp.BasicAuth(self.username, password=self.password)
        url = "https://{0}:{1}/nclu/v1/rpc".format(self.address, self.port)
        headers = {"Content-Type": "application/json"}

        try:
            async with aiohttp.ClientSession(
                    auth=auth,
                    timeout=timeout or self.cmd_timeout,
                    connector=aiohttp.TCPConnector(ssl=False),
            ) as session:
                for cmd in cmd_list:
                    data = {"cmd": cmd}
                    async with session.post(
                            url, json=data, headers=headers
                    ) as response:
                        result.append(
                            {
                                "status": response.status,
                                "timestamp": int(datetime.now(tz=timezone.utc)
                                                 .timestamp() * 1000),
                                "cmd": cmd,
                                "devtype": self.devtype,
                                "namespace": self.nsname,
                                "hostname": self.hostname,
                                "address": self.address,
                                "data": await response.text(),
                            }
                        )
        except Exception as e:
            self.current_exception = e
            result.append(self._create_error(cmd_list))
            self.logger.error("ERROR: (REST) Unable to communicate with node "
                              "%s:%d due to %s", self.address, self.port,
                              e)

        await service_callback(result, cb_token)


class LinuxNode(CumulusNode):
    '''Linux server node'''


class IosXRNode(Node):
    '''IOSXR Node specific implementation'''

    async def _init_rest(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat='json', timeout=None):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data(self):
        """Fill in the boot time of the node by executing certain cmds"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["show version", "show run hostname"],
                             None, 'text')

    async def _init_ssh(self, init_dev_data=True,
                        use_lock: bool = True) -> None:
        '''Need to start a neverending process to keep persistent ssh

        IOS XR's ssh is fragile and archaic. It doesn't support sending
        multiple commands over a single SSH connection as most other devices
        do. I suspect this may not be the only one. The bug is mentioned in
        https://github.com/ronf/asyncssh/issues/241. To overcome this issue,
        we wait in a loop for things to succeed. There's no point in continuing
        if this doesn't succeed. Maybe better to abort after a fixed number
        of retries to enable things like run-once=gather to work.
        '''
        backoff_period = 1
        if use_lock:
            await self.ssh_ready.acquire()

        while not self._conn:

            if not self._retry:
                break

            await super()._init_ssh(init_dev_data=False, use_lock=False)

            if self.is_connected:
                break

            await asyncio.sleep(backoff_period)
            backoff_period *= 2
            backoff_period = min(backoff_period, 120)

        if self.is_connected and not self._long_proc:
            try:
                self._long_proc = await self._conn.create_process(
                    'run tail -s 3600 -f /etc/version', stdout=DEVNULL,
                    stderr=DEVNULL)
                self.logger.info(
                    f'Persistent SSH present for {self.hostname}')
                if init_dev_data:
                    await self._fetch_init_dev_data()
            except Exception:
                self._conn = self._long_proc = None

        if use_lock:
            self.ssh_ready.release()

    async def _parse_init_dev_data(self, output, cb_token) -> None:
        '''Parse the version for uptime and hostname'''
        if output[0]["status"] == 0:
            data = output[0]['data']
            timestr = re.search(r'uptime is (.*)\n', data)
            if timestr:
                self.bootupTimestamp = int(datetime.utcfromtimestamp(
                    parse(timestr.group(1)).timestamp()).timestamp()*1000)
            else:
                self.logger.error(
                    f'Cannot parse uptime from {self.address}:{self.port}')
                self.bootupTimestamp = -1

            self._extract_nos_version(data)

        if (len(output) > 1) and (output[1]["status"] == 0):
            data = output[1]['data']
            hostname = re.search(r'hostname (\S+)', data.strip())
            if hostname:
                self.hostname = hostname.group(1)
                self.logger.error(f'set hostname of {self.address}:{self.port}'
                                  f' to {hostname.group(1)}')

    def _extract_nos_version(self, data) -> None:
        match = re.search(r'Version\s+:\s+ (\S+)', data)
        if match:
            self.version = match.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"


class IosXENode(Node):
    '''IOS-XE Node-sepcific telemetry gather implementation'''

    async def _init_rest(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat='json', timeout=None):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data(self):
        """Fill in the boot time of the node by executing certain cmds"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["show version"], None, 'text')

    async def _init_ssh(self, init_dev_data=True,
                        use_lock: bool = True) -> None:
        '''Need to start an interactive session for XE

        Many IOSXE devices cannot accept commands as fast as we can fire them.
        Furthermore, many TACACS servers also collapse under the load of
        multiple authemtication requests because each SSH session needs to be
        authenticated, not just the main SSH connection.
        '''
        backoff_period = 1
        self.WAITFOR = r'.*[>#]\s*$'

        self.logger.info(
            f'Trying to reconnect via SSH for {self.hostname}')

        if not self._retry or (self._conn and self._stdin):
            return

        if use_lock:
            await self.ssh_ready.acquire()

        while not self.is_connected:
            # Don't release rel lock here
            await super()._init_ssh(init_dev_data=False, use_lock=False)

            if self.is_connected:
                self.logger.info(
                    f'Reconnect succeeded via SSH for {self.hostname}')
                break

            await asyncio.sleep(backoff_period)
            backoff_period *= 2
            backoff_period = min(backoff_period, 120)

        if self.is_connected and not self._stdin:
            self.logger.info(
                f'Trying to create Persistent SSH for {self.hostname}')
            try:
                self._stdin, self._stdout, self._stderr = \
                    await self._conn.open_session(term_type='xterm')
                self.logger.info(
                    f'Persistent SSH created for {self.hostname}')

                output = await self.wait_for_prompt()
                if output.strip().endswith('>'):
                    if await self._handle_privilege_escalation() == -1:
                        await self._close_connection()
                        self._conn = None
                        self._stdin = None
                        self._retry = False  # No retry if escalation fails
                        if use_lock:
                            self.ssh_ready.release()
                        return
            except Exception as e:
                self.current_exception = e
                self.logger.error('Unable to create persistent SSH session'
                                  f' for {self.hostname} due to {str(e)}')
                self._conn = None
                self._stdin = None
                if use_lock:
                    self.ssh_ready.release()
                return

            # Set the terminal length to 0 to avoid paging
            self._stdin.write('terminal length 0\n')
            output = await self._stdout.readuntil(self.WAITFOR)
            if init_dev_data:
                await self._fetch_init_dev_data()

        if use_lock:
            self.ssh_ready.release()
        return

    async def _handle_privilege_escalation(self) -> int:
        '''Escalata privilege if necessary

        Returns 0 on success, -1 otherwise'''

        self.logger.info(
            f'Privilege escalation required for {self.hostname}')
        self._stdin.write('enable\n')
        output = await self.wait_for_prompt(r'Password:\s*')
        if self.enable_password:
            self._stdin.write(self.enable_password + '\n')
        else:
            self._stdin.write(self.password + '\n')

        output = await self.wait_for_prompt()
        if (output in ['suzieq timeout', 'Password:'] or
                output.strip().endswith('>')):
            self.logger.error(
                f'Privilege escalation failed for {self.hostname}'
                ', Aborting connection')
            return -1

        self.logger.info(f'Privilege escalation succeeded for {self.hostname}')
        return 0

    async def wait_for_prompt(self, prompt: str = None,
                              timeout: int = 90) -> str:
        """Wait for specified prompt upto timeout duration

        Since we're waiting for a prompt, we want to not wait forever.
        asyncssh's readuntil doesn't take a timeout parameter as of
        2.9.0. So, instead of adding a asyncio.waitfor everywhere, we
        just call this routine. By default, we wait for 90s

        Args:
            prompt[str]: The prompt string to wait for
            timeout[int]: How long to wait in secs
        Returns:
            the output data or 'timeout'
        """
        if prompt is None:
            prompt = self.WAITFOR
        coro = self._stdout.readuntil(prompt)
        try:
            output = await asyncio.wait_for(coro, timeout=timeout)
            return output
        except asyncio.TimeoutError:
            self.current_exception = asyncio.TimeoutError
            self.logger.error(f'{self.address}.{self.port} '
                              'Timed out waiting for expected prompt')
            # Return something that won't ever be in real output
            return 'suzieq timeout'

    async def _parse_init_dev_data(self, output, cb_token) -> None:
        '''Parse the version for uptime and hostname'''
        if not isinstance(output, list):
            # In some errors, the output returned is not a list
            self.bootupTimestamp = -1
            return

        if output[0]["status"] == 0:
            data = output[0]['data']
            hostupstr = re.search(r'(\S+)\s+uptime is (.*)\n', data)
            if hostupstr:
                self._set_hostname(hostupstr.group(1))
                timestr = hostupstr.group(2)
                self.bootupTimestamp = int(datetime.utcfromtimestamp(
                    parse(timestr).timestamp()).timestamp()*1000)
            else:
                self.logger.error(
                    f'Cannot parse uptime from {self.address}:{self.port}')
                self.bootupTimestamp = -1

            self._extract_nos_version(data)

    async def _ssh_gather(self, service_callback, cmd_list, cb_token, oformat,
                          timeout, only_one=False):
        """Run ssh for cmd in cmdlist and place output on service callback
           This is different from IOSXE to avoid reinit node info each time
        """

        result = []
        if cmd_list is None:
            await service_callback({}, cb_token)
            return

        if not self._conn or not self._stdin:
            await self._init_ssh()

        if not self._conn or not self._stdin:
            self.logger.error(f'Not connected to {self.address}:{self.port}')
            await service_callback({}, cb_token)
            return

        timeout = timeout or self.cmd_timeout
        for cmd in cmd_list:
            try:
                if self.slow_host:
                    await asyncio.sleep(IOS_SLEEP_BET_CMDS)
                self._stdin.write(cmd + '\n')
                output = await self.wait_for_prompt()
                if 'Invalid input detected' in output:
                    status = -1
                elif 'suzieq timeout' in output:
                    status = HTTPStatus.REQUEST_TIMEOUT
                else:
                    status = 0
                if isinstance(cb_token, RsltToken):
                    cb_token.node_token = self.bootupTimestamp
                result.append(self._create_result(cmd, status, output))
                continue
            except Exception as e:
                self.current_exception = e
                result.append(self._create_error(cmd))
                if not isinstance(e, asyncio.TimeoutError):
                    self.logger.error(
                        f"Unable to connect to {self.hostname} for {cmd} "
                        f"due to {e}")
                    await self._close_connection()
                else:
                    self.logger.error(
                        f"Unable to connect to {self.hostname} {cmd} "
                        "due to timeout")
                break

        await service_callback(result, cb_token)

    def _extract_nos_version(self, data: str) -> None:
        match = re.search(r', Version\s+([^ ,]+)', data)
        if match:
            self.version = match.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"


class IOSNode(IosXENode):
    '''Classic IOS Node-specific implementation'''

    async def _init_rest(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat='json', timeout=None):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')


class JunosNode(Node):
    '''Juniper's Junos node-specific implementation'''

    async def _init_rest(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat='json', timeout=None):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data(self):
        """Fill in the boot time of the node by running requisite cmd"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["show system uptime|display json",
                              "show version"], None, 'mixed')

    async def _parse_init_dev_data(self, output, cb_token) -> None:
        """Parse the uptime command output"""
        if output[0]["status"] == 0:
            data = output[0]["data"]
            try:
                jdata = json.loads(data.replace('\n', '').strip())
                if self.devtype not in ["junos-mx", "junos-qfx10k"]:
                    jdata = (jdata['multi-routing-engine-results'][0]
                             ['multi-routing-engine-item'][0])

                timestr = (jdata['system-uptime-information'][0]
                           ['system-booted-time'][0]['time-length'][0]
                           ['attributes'])
            except Exception:
                self.logger.warning(
                    f'Unable to parse junos boot time from {data}')
                timestr = '{"junos:seconds": "0"}'
            self.bootupTimestamp = (get_timestamp_from_junos_time(
                timestr, output[0]['timestamp']/1000)/1000)

        if (len(output) > 1) and (output[1]["status"] == 0):
            data = output[1]["data"]
            hmatch = re.search(r'\nHostname:\s+(\S+)\n', data)
            if hmatch:
                self._set_hostname(hmatch.group(1))

            self._extract_nos_version(data)

    def _extract_nos_version(self, data) -> None:
        """Extract the version from the output of show version"""
        match = re.search(r'Junos: (\S+)', data)
        if match:
            self.version = match.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"


class NxosNode(Node):
    '''Cisco's NXOS Node-specific implementation'''

    async def _init_rest(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None):
        '''Gather data for service via device REST API'''
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data(self):
        """Fill in the boot time of the node by running requisite cmd"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["show version|json", "show hostname"], None,
                             'mixed')

    async def _parse_init_dev_data(self, output, cb_token) -> None:
        """Parse the uptime command output"""

        hostname = ''
        if output[0]["status"] == 0:
            data = json.loads(output[0]["data"])
            upsecs = (24*3600*int(data.get('kern_uptm_days', 0)) +
                      3600*int(data.get('kern_uptm_hrs', 0)) +
                      60*int(data.get('kern_uptm_mins', 0)) +
                      int(data.get('kern_uptm_secs', 0)))
            if upsecs:
                self.bootupTimestamp = int(int(time.time()*1000)
                                           - float(upsecs)*1000)
            self.version = data.get('nxos_ver_str', '')
            if not self.version:
                self.logger.error(
                    f'Cannot extract version from {self.address}:{self.port}')

        if len(output) > 1:
            if output[1]["status"] == 0:
                hostname = output[1]["data"].strip()
        else:
            if output[0]['hostname'] != output[0]['address']:
                hostname = output[0]['hostname']

        if hostname:
            self._set_hostname(hostname)

    def _extract_nos_version(self, data: str) -> None:
        match = re.search(r'NXOS:\s+version\s+(\S+)', data)
        if match:
            self.version = match.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"


class SonicNode(Node):
    '''SONiC Node-specific implementtaion'''

    async def _init_rest(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None):
        '''Gather data for service via device REST API'''
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data(self):
        """Fill in the boot time of the node by running requisite cmd"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["cat /proc/uptime", "hostname", "show version"],
                             None, 'text')

    async def _parse_init_dev_data(self, output, cb_token) -> None:
        """Parse the uptime command output"""

        if output[0]["status"] == 0:
            upsecs = output[0]["data"].split()[0]
            self.bootupTimestamp = int(int(time.time()*1000)
                                       - float(upsecs)*1000)
        if (len(output) > 1) and (output[1]["status"] == 0):
            self.hostname = output[1]["data"].strip()
        if (len(output) > 2) and (output[2]["status"] == 0):
            self._extract_nos_version(output[1]["data"])

    def _extract_nos_version(self, data: str) -> None:
        match = re.search(r'Version:\s+SONiC-OS-([^-]+)', data)
        if match:
            self.version = match.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"


class PanosNode(Node):
    '''Node object representing access to a Palo Alto Networks FW'''

    async def _fetch_init_dev_data(self):
        try:
            res = []
            # temporary hack to detect device info using ssh
            async with asyncssh.connect(
                    self.address, port=22, username=self.username,
                    password=self.password, known_hosts=None) as conn:
                async with conn.create_process() as process:
                    process.stdin.write("show system info\n")
                    recv_chars = False
                    output = ""
                    if not recv_chars:
                        output += await process.stdout.read(1)
                    try:
                        await asyncio.wait_for(
                            process.wait_closed(), timeout=0.01)
                    except asyncio.TimeoutError:
                        pass

                    stdout, _ = process.collect_output()
                    output += stdout
                    res = [{
                        "status": 0,
                        "data": output}]
                    await self._parse_init_dev_data(res, None)
            self._session = aiohttp.ClientSession(
                conn_timeout=self.connect_timeout,
                connector=aiohttp.TCPConnector(ssl=False),
            )
            if self.api_key is None:
                await self.get_api_key()
        except Exception:
            pass

    async def get_api_key(self):
        """Authenticate to get the api key needed in all cmd requests"""
        url = f"https://{self.address}:{self.port}/api/?type=keygen&user=" \
            f"{self.username}&password={self.password}"
        async with self._session.get(url, timeout=self.connect_timeout) \
                as response:
            status, xml = response.status, await response.text()
            if status == 200:
                data = xmltodict.parse(xml)
                self.api_key = data["response"]["result"]["key"]
            # need to manage errors

    async def _parse_init_dev_data(self, output, cb_token) -> None:
        """Parse the uptime command output"""
        if output[0]["status"] == 0:
            data = output[0]["data"]
            # extract uptime
            match = re.search(
                r'uptime:\s+(\d+)\sdays,\s(\d+):(\d+):(\d+)', data)

            if match:
                days = match.group(1).strip()
                hours = match.group(2).strip()
                minutes = match.group(3).strip()
                seconds = match.group(4).strip()
                upsecs = 86400 * int(days) + 3600 * int(hours) + \
                    60 * int(minutes) + int(seconds)
                self.bootupTimestamp = int(
                    int(time.time()*1000) - float(upsecs)*1000)
            else:
                self.logger.warning(
                    f'Cannot parse uptime from {self.address}:{self.port}')

            # extract hostname
            hmatch = re.search(r'hostname:\s+(\S+)\n', data)
            if hmatch:
                self.hostname = hmatch.group(1).strip()
            else:
                self.logger.warning(
                    f'Cannot parse hostname from {self.address}:{self.port}')

            self._extract_nos_version(data)

    def _extract_nos_version(self, data: str) -> None:
        vmatch = re.search(r'sw-version:\s+(\S+)', data)
        if vmatch:
            self.version = vmatch.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"

    async def _init_rest(self):
        # In case of PANOS, getting here means REST is up
        if not self._session:
            try:
                self._session = aiohttp.ClientSession(
                    conn_timeout=self.connect_timeout,
                    connector=aiohttp.TCPConnector(ssl=False),
                )
                if self.api_key is None:
                    await self.get_api_key()
            except Exception as e:
                self.logger.error(
                    f'Unable to connect to {self.address}:{self.port}, '
                    f'error: {str(e)}')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None):

        result = []
        if not cmd_list:
            return result

        timeout = timeout or self.connect_timeout

        now = int(datetime.now(tz=timezone.utc).timestamp() * 1000)

        port = self.port or 443

        url = f"https://{self.address}:{port}/api/"

        status = 200  # status OK

        if not self._session:
            for cmd in cmd_list:
                result.append(self._create_error(cmd))
            await service_callback(result, cb_token)
            return

        try:
            for cmd in cmd_list:
                url_cmd = f"{url}?type=op&cmd={cmd}&key={self.api_key}"
                async with self._session.get(
                        url_cmd, timeout=timeout) as response:
                    status, xml = response.status, await response.text()
                    json_out = json.dumps(
                        xmltodict.parse(xml))

                    result.append({
                        "status": status,
                        "timestamp": now,
                        "cmd": cmd,
                        "devtype": self.devtype,
                        "namespace": self.nsname,
                        "hostname": self.hostname,
                        "address": self.address,
                        "data": json_out,
                    })
        except Exception as e:
            self.current_exception = e
            for cmd in cmd_list:
                result.append(self._create_error(cmd))
            self.logger.error("ERROR: (REST) Unable to communicate with node "
                              f"{self.address}:{self.port} due to {str(e)}")

        await service_callback(result, cb_token)



from nubia import command
import pandas as pd

from suzieq.cli.nubia_patch import argument
from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.fs import FsObj


@command("fs", help="Act on File System data")
@argument("used_percent", description="must be of the form "
          "[<|<=|>=|>|!]value. Eg: '<=20'")
@argument("mountPoint", description="Mount point(s), space separated")
class FsCmd(SqCommand):
    """Filesystem information such as total disk space, filesystems etc"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = " ",
            columns: str = "default",
            used_percent: str = '',
            mountPoint: str = ''
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=FsObj,
        )
        self.lvars = {
            'usedPercent': used_percent,
            'mountPoint': mountPoint.split()
        }

    @command("show")
    def show(self):
        """Show File System info
        """
        # Get the default display field names
        used_percent = self.lvars.get('usedPercent')
        if used_percent and not any(used_percent.startswith(x)
                                    for x in ['<=', '>=', '<', '>', '!']):
            try:
                int(used_percent)
            except ValueError:
                df = pd.DataFrame(
                    {'error': ['ERROR invalid used-percent operation']})
                return self._gen_output(df)

        return super().show()

    @command("summarize")
    def summarize(self, **kwargs):
        """
        Summarize the filesystem/storage info
        """
        df = pd.DataFrame({'error': ['Summarize not yet supported for FS']})
        return self._gen_output(df)



from nubia import command

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.mlag import MlagObj


@command('mlag', help="Act on mlag data")
class MlagCmd(SqCommand):
    """Multichassis LAG information (includes variants such as NXOX vPC)"""

    def __init__(self, engine: str = '', hostname: str = '',
                 start_time: str = '', end_time: str = '',
                 view: str = '', namespace: str = '',
                 query_str: str = ' ',
                 format: str = "", columns: str = 'default') -> None:
        # pylint: disable=redefined-builtin
        super().__init__(engine=engine, hostname=hostname,
                         start_time=start_time, end_time=end_time,
                         view=view, namespace=namespace,
                         format=format, query_str=query_str, columns=columns,
                         sqobj=MlagObj)



import os
import sys
import subprocess
import argparse
from importlib.util import find_spec
from pathlib import Path

from colorama import init, Fore, Style
from suzieq.shared.utils import load_sq_config, print_version


def gui_main(*args):
    '''Kicks things off'''
    if not args:
        args = sys.argv

    parser = argparse.ArgumentParser(args)
    parser.add_argument(
        "-c",
        "--config",
        type=str, help="alternate config file",
        default=None
    )
    parser.add_argument(
        "-F",
        "--framework",
        type=str, help="Name of GUI framework to use",
        choices=['streamlit'],
        default='streamlit'
    )
    parser.add_argument(
        "--version",
        "-V",
        help="print Suzieq version",
        default=False, action='store_true',
    )
    parser.add_argument(
        "--port",
        "-p",
        help="http port to connect to",
        default='8501',
    )
    userargs = parser.parse_args()

    if userargs.version:
        print_version()
        sys.exit(0)

    spec = find_spec('suzieq.gui')
    if spec:
        if userargs.framework == 'streamlit':
            thisprog = Path(spec.loader.path).parent / \
                'stlit' / 'suzieq-gui.py'
            thisprog = str(thisprog)

    # validate the config file before starting the gui
    load_sq_config(config_file=userargs.config)

    # write the sq_config file path inside an environment variable
    os.environ['SQ_GUI_CONFIG_FILE'] = userargs.config or ''

    init()
    with subprocess.Popen(['streamlit', 'run', thisprog,
                           '--server.port', userargs.port,
                           '--theme.base', 'light']) as p:
        print(Fore.CYAN + Style.BRIGHT +
              "\n  Starting Suzieq GUI" + Style.RESET_ALL)
        p.communicate()


if __name__ == '__main__':
    gui_main()



"""
This module contains the Poller logic
"""

import argparse
import asyncio
import logging
import os
import signal
from typing import Dict, Type

from suzieq.poller.worker.inventory.inventory import Inventory
from suzieq.poller.worker.services.service_manager import ServiceManager
from suzieq.poller.worker.writers.output_worker_manager \
    import OutputWorkerManager
from suzieq.shared.exceptions import SqPollerConfError

logger = logging.getLogger(__name__)


class Worker:
    """Worker is the object in charge of coordinating services, nodes and
    output worker tasks, in order to pull the data from the devices configured
    in the device inventory.
    """

    DEFAULT_INVENTORY = 'static'

    def __init__(self, userargs, cfg):
        self._validate_worker_args(userargs, cfg)

        # Set the worker id
        self.worker_id = userargs.worker_id

        # Setup poller tasks list
        self.waiting_tasks = []
        self.waiting_tasks_lock = asyncio.Lock()

        # Setup poller writers

        # TODO: At the moment:
        # output_dir: is the directory used by the gather method
        # data_dir: is the directory used by parquet
        # we need a way to define the settings
        # for each type of output worker
        self.output_args = {
            'output_dir': userargs.output_dir,
            'data_dir': cfg.get('data-directory')
        }

        if userargs.run_once in ['gather', 'process']:
            userargs.outputs = ['gather']
        self.output_manager = OutputWorkerManager(userargs.outputs,
                                                  self.output_args)
        self.output_queue = self.output_manager.output_queue

        # Initialize service manager
        service_dir = cfg['service-directory']
        svc_schema_dir = cfg.get('schema-directory', None)
        default_svc_period = cfg.get('poller', {}).get('period', 15)
        run_mode = userargs.run_once or 'forever'
        svc_manager_args = {
            'service_only': userargs.service_only,
            'exclude_services': userargs.exclude_services,
            'outputs': userargs.outputs
        }
        self.service_manager = ServiceManager(self._add_worker_tasks,
                                              service_dir,
                                              svc_schema_dir,
                                              self.output_queue,
                                              run_mode,
                                              cfg,
                                              default_svc_period,
                                              **svc_manager_args)

        # Init the node inventory object
        self.inventory = self._init_inventory(userargs, cfg)

    async def init_worker(self):
        """Initialize the worker, instantiating the services and setting up
        the connection with the nodes. This function should be called only
        at the beginning before calling run().
        """

        logger.info('Initializing poller worker')

        # When the poller receives a termination signal, we would like
        # to gracefully terminate all the tasks, i.e. closing all the
        # connections with nodes.
        loop = asyncio.get_event_loop()
        for s in [signal.SIGTERM, signal.SIGINT]:
            loop.add_signal_handler(
                s, lambda s=s: asyncio.create_task(self._stop()))

        init_tasks = []
        init_tasks.append(self.inventory.build_inventory())
        init_tasks.append(self.service_manager.init_services())

        nodes, services = await asyncio.gather(*init_tasks)

        if not nodes:
            # Get the logger filename
            root_logger = logging.getLogger()
            log_filename = ""
            for h in root_logger.handlers:
                if hasattr(h, 'baseFilename'):
                    log_filename = h.baseFilename
                    break

            raise SqPollerConfError(
                f'No nodes could be polled. Check {log_filename} for errors '
                'in connecting'
            )
        if not services:
            raise SqPollerConfError('No services candidate for execution')

    async def run(self):
        """Start polling the devices.
        Before running this function the worker should be initialized.
        """

        # Add the node list in the services
        await self.service_manager.set_nodes(self.inventory.get_node_callq())

        logger.info('Suzieq Started')

        # Schedule the tasks to run
        await self.inventory.schedule_nodes_run()
        await self.service_manager.schedule_services_run()
        await self._add_worker_tasks(
            [self.output_manager.run_output_workers()])

        try:
            # The logic below of handling the writer worker task separately
            # is to ensure we can terminate properly when all the other
            # tasks have finished as in the case of using file input
            # instead of SSH
            tasks = await self._pop_waiting_worker_tasks()
            while tasks:
                try:
                    _, pending = await asyncio.wait(
                        tasks, return_when=asyncio.FIRST_COMPLETED)
                    tasks = list(pending)
                    running_svcs = self.service_manager.running_services
                    # pylint: disable=protected-access
                    if tasks and any(i._coro in running_svcs
                                     for i in tasks):
                        continue

                    break
                except asyncio.CancelledError:
                    break
        except asyncio.CancelledError:
            logger.warning('Received terminate signal. Terminating...')

    async def _add_worker_tasks(self, tasks):
        """Add new tasks to be executed in the poller worker run loop."""

        await self.waiting_tasks_lock.acquire()
        self.waiting_tasks += tasks
        self.waiting_tasks_lock.release()

    def _get_inventory_plugins(self) -> Dict[str, Type]:
        """Get all types of inventory it is possible to instantiate

        Returns:
            Dict[str, Type]: a dictionary containing the name of the inventory
                type and its class
        """
        return Inventory.get_plugins()

    def _init_inventory(self, userargs: argparse.Namespace, cfg: Dict,
                        addnl_args: Dict = None):
        """Initialize the Inventory object, in charge of retrieving the
        list of nodes this worker is in charge to poll

        Args:
            userargs (argparse.Namespace): the cli arguments
            cfg (Dict): the content of the SuzieQ config file
            addnl_args (Dict): additional arguments to pass to the init
                function of the Inventory class

        Raises:
            SqPollerConfError: raised if a wrong configuration is provided
        """

        # Define the dictionary with the settings
        # for any kind of inventory source
        connect_timeout = cfg.get('poller', {}).get('connect-timeout', 15)
        inventory_args = {
            'connect_timeout': connect_timeout,
            'ssh_config_file': userargs.ssh_config_file,
        }

        if addnl_args:
            inventory_args.update(addnl_args)

        # Retrieve the specific inventory source to use
        inv_types = self._get_inventory_plugins()

        inventory_class = None
        source_args = {}

        if userargs.input_dir:
            # 'dir' is not a real inventory source
            # we need to override the Inventory class
            # in order to simulate nodes providing the data
            # inside the specified input directory.
            inventory_class = inv_types['dir']
            source_args = {'input_dir': userargs.input_dir}
        else:
            mgr_cfg = cfg.get('poller', {}).get('manager', {})
            type_to_use = mgr_cfg.get('type', self.DEFAULT_INVENTORY)
            inventory_class = inv_types.get(type_to_use)
            if not inventory_class:
                raise SqPollerConfError(f'No inventory {type_to_use} found')
            source_args = {
                **mgr_cfg,
                'worker-id': self.worker_id
            }

        return inventory_class(self._add_worker_tasks,
                               **source_args,
                               **inventory_args)

    async def _pop_waiting_worker_tasks(self):
        """Empty the list of tasks to be added in the run loop
        and return its content.
        """

        # Since the function is asynchronous and we need to
        # read the content of the task list and, at the end, empty
        # it, we need to handle concurrency. Otherwise we risk to loose
        # all the tasks added after the list has been read, but before
        # the list emptying.
        await self.waiting_tasks_lock.acquire()
        poller_tasks = self.waiting_tasks
        self.waiting_tasks = []
        self.waiting_tasks_lock.release()

        return poller_tasks

    async def _stop(self):
        """Stop the worker"""

        tasks = [t for t in asyncio.all_tasks()
                 if t is not asyncio.current_task()]

        for task in tasks:
            task.cancel()

    def _validate_worker_args(self, userargs: Dict, _):
        """Validate the arguments and the configuration passed to the poller.
        The function produces a SqPollerConfError exception if there is
        something wrong in the configuration.

        Args:
            userargs (Dict): Dictionary containing the arguments passed to the
                poller
            cfg (Dict): The content of the Suzieq configuration file

        Raises:
            SqPollerConfError: raise when the configuration is not valid
        """

        if userargs.ssh_config_file:
            if not os.access(userargs.ssh_config_file, os.F_OK):
                raise SqPollerConfError(
                    f'Unable to read ssh config in {userargs.ssh_config_file}'
                )
            ssh_config_file = os.path.expanduser(userargs.ssh_config_file)
            if (os.stat(
                    os.path.dirname(
                        ssh_config_file)).st_mode | 0o40700 != 0o40700):
                raise SqPollerConfError(
                    'ssh directory has wrong permissions, must be 0700'
                )



import pandas as pd
from pandas.core.dtypes.dtypes import DatetimeTZDtype

from suzieq.sqobjects.basicobj import SqObject
from suzieq.shared.utils import humanize_timestamp


class BgpObj(SqObject):
    '''The object providing access to the bgp table'''

    def __init__(self, **kwargs):
        super().__init__(table='bgp', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'state',
                                'vrf', 'peer', 'asn', 'afiSafi', 'query_str']
        self._valid_arg_vals = {
            'state': ['Established', 'NotEstd', 'dynamic', ''],
            'result': ['all', 'pass', 'fail'],
        }
        self._valid_assert_args = self._valid_get_args + ['result']

    def humanize_fields(self, df: pd.DataFrame, _=None) -> pd.DataFrame:
        '''Humanize the timestamp and boot time fields'''
        if df.empty:
            return df

        if 'estdTime' in df.columns:
            if not isinstance(df.estdTime.dtype, DatetimeTZDtype):
                df['estdTime'] = humanize_timestamp(
                    df.estdTime, self.cfg.get('analyzer', {})
                    .get('timezone', None))

        return super().humanize_fields(df)



import sys

from nubia import context
from nubia import exceptions
from nubia import eventbus

from suzieq.shared.utils import load_sq_config, print_version
from suzieq.shared.schema import Schema
from suzieq.shared.context import SqContext


class NubiaSuzieqContext(context.Context):
    '''Suzieq Nubia context setup on CLI startup'''

    def __init__(self):
        self.ctxt = None
        super().__init__()

    def on_connected(self, *args, **kwargs):
        if self._args.V:
            print_version()
            sys.exit(0)
        if self._args.config:
            cfg = load_sq_config(validate=True,
                                 config_file=self._args.config)
        else:
            cfg = load_sq_config(validate=True)

        if not cfg:
            print('ERROR: No suzieq configuration found')
            print('Create a suzieq-cfg.yml under the homedir or current dir')
            print('OR pass a path to the config file via -c argument')
            sys.exit(1)
        self.ctxt = SqContext(cfg=cfg)
        self.ctxt.schemas = Schema(self.ctxt.cfg["schema-directory"])

    def on_cli(self, cmd, args):
        # dispatch the on connected message
        self.registry.dispatch_message(eventbus.Message.CONNECTED)

    def on_interactive(self, args):
        ret = self._registry.find_command("connect").run_cli(args)
        if ret:
            raise exceptions.CommandError("Failed starting interactive mode")
        # dispatch the on connected message

    def change_engine(self, engine: str):
        '''Change the backend engine'''
        if engine == self.ctxt.engine:
            return

        self.ctxt.engine = engine



from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.address import AddressObj


@argument("vrf",
          description="VRF(s), space separated")
@argument("type", description="Interface type(s), space separated")
@argument("ifname", description="Interface name(s), space separated")
@argument("ipvers",
          description="type of address, v4, v6 or l2",
          choices=["v4", "v6", "l2"])
@argument("address",
          description="Address(es), in quotes, space separated")
@argument("prefix",
          description=("Show all the addresses in this "
                       "subnet prefix (in quotes)"))
@command("address", help="Act on interface addresses")
class AddressCmd(SqCommand):
    """IP and MAC addresses associated with interfaces"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            columns: str = "default",
            query_str: str = ' ',
            vrf: str = '',
            type: str = '',     # pylint: disable=redefined-builtin
            ifname: str = '',
            ipvers: str = '',
            address: str = '',
            prefix: str = '',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            format=format,
            columns=columns,
            query_str=query_str,
            sqobj=AddressObj,
        )
        self.lvars = {
            'vrf': vrf.split(),
            'type': type.split(),
            'ifname': ifname.split(),
            'ipvers': ipvers,
            'address': address.split(),
            'prefix': prefix.split()
        }



import typing
import pandas as pd
from pandas.core.dtypes.dtypes import DatetimeTZDtype

from suzieq.shared.utils import load_sq_config, humanize_timestamp
from suzieq.shared.schema import Schema, SchemaForTable
from suzieq.engines import get_sqengine
from suzieq.shared.sq_plugin import SqPlugin
from suzieq.shared.context import SqContext


class SqObject(SqPlugin):
    '''The base class for accessing the backend independent of the engine'''

    def __init__(self, engine_name: str = '',
                 hostname: typing.List[str] = None,
                 start_time: str = '', end_time: str = '',
                 view: str = '', namespace: typing.List[str] = None,
                 columns: typing.List[str] = None,
                 context=None, table: str = '', config_file=None) -> None:

        if not context:
            self.ctxt = SqContext(cfg=load_sq_config(validate=True,
                                                     config_file=config_file),
                                  engine=engine_name)
            self.ctxt.schemas = Schema(self.ctxt.cfg["schema-directory"])
        else:
            self.ctxt = context
            if not self.ctxt.cfg:
                self.ctxt.cfg = load_sq_config(validate=True,
                                               config_file=config_file)
                self.ctxt.schemas = Schema(self.ctxt.cfg["schema-directory"])
            if not self.ctxt.engine:
                self.ctxt.engine = engine_name

        self._cfg = self.ctxt.cfg
        self._schema = SchemaForTable(table, self.ctxt.schemas)
        self._table = table
        self._sort_fields = self._schema.key_fields()
        self._convert_args = {}

        self.namespace = namespace or self.ctxt.namespace or []
        self.hostname = hostname or self.ctxt.hostname or []
        self.start_time = start_time or self.ctxt.start_time
        self.end_time = end_time or self.ctxt.end_time

        view = view or self.ctxt.view

        if self.start_time and self.end_time and not view:
            self.view = 'all'
        else:
            self.view = view or 'latest'

        self.columns = columns or ['default']
        self._unique_def_column = ['hostname']

        # Init the engine
        engine_to_use = engine_name if engine_name else self.ctxt.engine
        self.engine = get_sqengine(engine_to_use, self._table)(self)

        self.summarize_df = pd.DataFrame()

        self._addnl_filter = []
        self._valid_get_args = self._valid_assert_args = []
        self._valid_arg_vals = self._valid_find_args = []
        self._valid_summarize_args = []

    @property
    def all_schemas(self):
        '''Return the set of all schemas of tables supported'''
        return self.ctxt.schemas

    @property
    def schema(self):
        '''Return table-specific schema'''
        return self._schema

    @property
    def cfg(self):
        '''Return general suzieq config'''
        return self._cfg

    @property
    def table(self):
        '''Return the table served by this object'''
        return self._table

    @property
    def sort_fields(self):
        '''Return default list of fields to sort by'''
        return self._sort_fields

    @property
    def get_filter_fields(self):
        '''Return the set of fields in table that are specified as filters'''
        return [x for x in self._valid_get_args if x in self.schema.fields]

    @property
    def unique_default_column(self):
        '''Return the default unique column for this table'''
        return self._unique_def_column

    def _check_input_for_valid_args(self, good_arg_list, **kwargs,):
        '''Check that the provided set of kwargs is valid for the table'''
        if not good_arg_list:
            return

        # add standard args that are always
        good_arg_list = good_arg_list + (['namespace'])

        for arg in kwargs:
            if arg not in good_arg_list:
                raise AttributeError(
                    f"argument {arg} not supported for this command")

    def _check_input_for_valid_vals(self, good_arg_val_list, **kwargs):
        '''Check if the input is valid for the arg, if possible'''

        if not good_arg_val_list:
            return

        for arg, val in kwargs.items():
            if arg not in good_arg_val_list:
                continue

            if not isinstance(val, list):
                chkval = [val]
            else:
                chkval = val
            for v in chkval:
                if v not in good_arg_val_list[arg]:
                    raise ValueError(
                        f"invalid value {val} for argument {arg}")

    def validate_get_input(self, **kwargs):
        '''Validate the values of the get function'''
        self._check_input_for_valid_args(
            self._valid_get_args + ['columns'], **kwargs)
        self._check_input_for_valid_vals(self._valid_arg_vals, **kwargs)

    def validate_assert_input(self, **kwargs):
        '''Validate the values of the assert function'''
        self._check_input_for_valid_args(self._valid_assert_args, **kwargs)

    def validate_summarize_input(self, **kwargs):
        '''Validate the values of the summarize function'''
        self._check_input_for_valid_args(self._valid_get_args, **kwargs)

    def validate_columns(self, columns: typing.List[str]) -> bool:
        """Validate that the provided columns are valid for the table

        Args:
            columns (List[str]): list of columns

        Returns:
            bool: True if columns are valid
        Raises:
            ValueError: if columns are invalid
        """

        if columns in [['default'], ['*']]:
            return True

        table_schema = SchemaForTable(self._table, self.all_schemas)
        invalid_columns = [x for x in columns if x not in table_schema.fields]
        if invalid_columns:
            raise ValueError(f"Invalid columns specified: {invalid_columns}")
        return True

    def get(self, **kwargs) -> pd.DataFrame:
        '''Return the data for this table given a set of attributes'''
        if not self._table:
            raise NotImplementedError

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        if self._addnl_filter:
            kwargs['add_filter'] = self._addnl_filter

        # This raises exceptions if it fails
        try:
            self.validate_get_input(**kwargs)
        except (AttributeError, ValueError) as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        if 'columns' not in kwargs:
            kwargs['columns'] = self.columns or ['default']

        # This raises ValueError if it fails
        self.validate_columns(kwargs.get('columns', []))

        for k, v in self._convert_args.items():
            if v and k in kwargs:
                val = kwargs[k]
                newval = []
                if isinstance(val, list):
                    for ele in val:
                        ele = v(ele)
                        newval.append(ele)
                    kwargs[k] = newval
                elif isinstance(val, str):
                    kwargs[k] = v(val)

        return self.engine.get(**kwargs)

    def summarize(self, **kwargs) -> pd.DataFrame:
        '''Summarize the data from specific table'''
        if self.columns != ["default"]:
            self.summarize_df = pd.DataFrame(
                {'error':
                 ['ERROR: You cannot specify columns with summarize']})
            return self.summarize_df

        if not self._table:
            raise NotImplementedError

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        self.validate_summarize_input(**kwargs)

        return self.engine.summarize(**kwargs)

    def unique(self, **kwargs) -> pd.DataFrame:
        '''Identify unique values and value counts for a column in table'''
        if not self._table:
            raise NotImplementedError

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        columns = kwargs.pop('columns', self.columns)
        count = kwargs.pop('count', 'False')

        if columns is None or columns == ['default']:
            columns = self._unique_def_column

        if len(columns) > 1 or columns == ['*']:
            raise ValueError('Specify a single column with unique')

        # This raises ValueError if it fails
        self.validate_columns(columns)
        self._check_input_for_valid_vals(self._valid_arg_vals, **kwargs)
        return self.engine.unique(**kwargs, count=str(count), columns=columns)

    def aver(self, **kwargs):
        '''Assert one or more checks on table'''
        if self._valid_assert_args:
            return self._assert_if_supported(**kwargs)

        raise NotImplementedError

    def top(self, what: str = '', count: int = 5, reverse: bool = False,
            **kwargs) -> pd.DataFrame:
        """Get the list of top/bottom entries of "what" field"""

        columns = kwargs.get('columns', ['default'])
        # This raises ValueError if it fails
        self.validate_columns(columns)

        if not what:
            raise ValueError('Must specify what field to get top for')
        # if self._valid_get_args:
        #     self._valid_get_args += ['what', 'n', 'reverse']
        # This raises exceptions if it fails
        try:
            self.validate_get_input(**kwargs)
        except (ValueError, AttributeError) as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        # This raises ValueError if it fails
        table_schema = SchemaForTable(self._table, self.all_schemas)
        if not self._field_exists(table_schema, what):
            raise ValueError(
                f"Field {what} does not exist in table {self.table}")

        columns = table_schema.get_display_fields(columns)

        ftype = table_schema.field(what).get('type', 'str')
        if ftype not in ['long', 'double', 'float', 'int', 'timestamp',
                         'timedelta64[s]']:
            return pd.DataFrame({'error':
                                 [f'{what} not numeric; top can be used with'
                                  f' numeric fields only']})

        return self.engine.top(what=what, count=int(count), reverse=reverse,
                               **kwargs)

    def describe(self, **kwargs):
        """Describes the fields for a given table"""

        table = kwargs.get('table', self.table)

        cols = kwargs.get('columns', ['default'])
        if cols not in [['default'], ['*']]:
            df = pd.DataFrame(
                {'error': ['ERROR: cannot specify column names for describe']})
            return df

        try:
            sch = SchemaForTable(table, self.all_schemas)
        except ValueError:
            sch = None
        if not sch:
            df = pd.DataFrame(
                {'error': [f'ERROR: incorrect table name {table}']})
            return df

        entries = [{'name': x['name'], 'type': x['type'],
                    'key': x.get('key', ''),
                    'display': x.get('display', ''),
                    'description': x.get('description', '')}
                   for x in sch.get_raw_schema()]
        df = pd.DataFrame.from_dict(entries).sort_values('name')

        query_str = kwargs.get('query_str', '')
        if query_str:
            return df.query(query_str).reset_index(drop=True)

        return df

    def get_table_info(self, **kwargs) -> pd.DataFrame:
        """Get some basic stats about the table from the database

        Args:
            kwargs: keyword args passed by caller, varies depending on table

        Returns:
            pd.DataFrame: A dataframe with the stats
        """
        # This raises ValueError if it fails
        self.validate_columns(kwargs.get('columns', ['default']))

        return self.engine.get_table_info(**kwargs)

    def humanize_fields(self, df: pd.DataFrame, _=None) -> pd.DataFrame:
        '''Humanize the fields for human consumption.

        Individual classes will implement the right transofmations. This
        routine is just a placeholder for all those with nothing to modify.
        '''
        if 'timestamp' in df.columns and not df.empty:
            if isinstance(df.timestamp.dtype, DatetimeTZDtype):
                return df
            df['timestamp'] = humanize_timestamp(df.timestamp,
                                                 self.cfg.get('analyzer', {})
                                                 .get('timezone', None))

        return df

    def _field_exists(self, table_schema: SchemaForTable, field: str) -> bool:
        """Check if a field exists in the schema

        Args:
            table_schema (SchemaForTable): The schema for the table
            field (str): the field name we're checking for

        Returns:
            bool: True if the field exists, False otherwise
        """
        return table_schema.field(field)

    def _assert_if_supported(self, **kwargs):
        '''Common sqobj routine for a table that supports asserts

           Do not call this routine directly
        '''

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')
        try:
            self.validate_assert_input(**kwargs)
        except AttributeError as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        if self.columns in [['*'], ['default']]:
            req_cols = None
        else:
            req_cols = self.schema.get_display_fields(self.columns)
            if not req_cols:
                # Till we add a schema object for assert columns,
                # this will have to do
                req_cols = self.columns

        df = self.engine.aver(**kwargs)
        if not df.empty and req_cols:

            req_col_set = set(req_cols)
            got_col_set = set(df.columns)
            diff_cols = req_col_set - got_col_set
            if diff_cols:
                return pd.DataFrame(
                    {'error': [f'columns {list(diff_cols)} not in dataframe']})

            if 'assert' not in req_cols:
                req_cols.append('assert')

            df = df[req_cols]

        return df



import re
import pandas as pd

from suzieq.sqobjects.basicobj import SqObject


class RoutesObj(SqObject):
    '''The object providing access to the routes table'''

    def __init__(self, **kwargs):
        super().__init__(table='routes', **kwargs)
        self._addnl_filter = 'metric != 4278198272'
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'prefix',
                                'vrf', 'protocol', 'prefixlen', 'ipvers',
                                'add_filter', 'address', 'query_str']
        self._unique_def_column = ['prefix']
        self._valid_summarize_args += ['vrf']

    def validate_get_input(self, **kwargs):
        '''Validate input to show'''
        if kwargs.get('prefixlen', ''):
            p_match = re.fullmatch(r'([<>][=]?|[!])?[ ]?([0-9]+)',
                                   kwargs['prefixlen'])
            if p_match:
                plen = int(p_match.group(2))
                operator = str(p_match.group(1))
            else:
                raise ValueError("Invalid prefixlen query: it must be of the "
                                 "form '[<|<=|>=|>|!] length' (i.e. '>= 24')")
            if (not (0 <= plen <= 128) or
               (plen == 128 and operator.startswith('>')) or
               (not plen and operator.startswith('<'))):
                raise ValueError("Invalid prefixlen: "
                                 "value should be between 0 and 128")

        return super().validate_get_input(**kwargs)

    def lpm(self, **kwargs) -> pd.DataFrame:
        '''Get the lpm for the given address'''
        if not kwargs.get("address", None):
            raise AttributeError('ip address is mandatory parameter')
        if isinstance(kwargs['address'], list):
            if len(kwargs['address']) > 1:
                raise AttributeError('Just one address at a time')
            kwargs['address'] = kwargs['address'][0]
        return self.engine.lpm(**kwargs)



from collections import namedtuple, OrderedDict
from inspect import isclass

from nubia.internal.helpers import (
    get_arg_spec,
    function_to_str,
    transform_name,
)


Argument = namedtuple(
    "Argument",
    "arg description type "
    "default_value_set default_value "
    "name extra_names positional choices",
)

Command = namedtuple("Command", "name help aliases exclusive_arguments")

FunctionInspection = namedtuple(
    "FunctionInspection", "arguments " "command subcommands"
)
_ArgDecoratorSpec = namedtuple(
    "_ArgDecoratorSpec", "arg name aliases description positional choices"
)

# pylint: disable=redefined-builtin


def argument(
    arg,
    type=None,
    description=None,
    name=None,
    aliases=None,
    positional=False,
    choices=None,
):
    """
    Annotation decorator to specify metadata for an argument

    Check the module documentation for more info and tests.py in this module
    for usage examples
    """

    def decorator(function):
        # Following makes interactive really slow. (T20898480)
        # This should be revisited in T20899641
        #  if (description is not None and \
        #       arg is not None and type is not None):
        #      append_doc(function, arg, type, description)

        fn_specs = get_arg_spec(function)
        args = fn_specs.args or []
        if arg not in args and not fn_specs.varkw:
            raise NameError(
                "Argument {} does not exist in function {}".format(
                    arg, function_to_str(function)
                )
            )

        # init structures to store decorator data if not present
        _init_attr(function, "__annotations__", OrderedDict())
        _init_attr(function, "__arguments_decorator_specs", {})

        # Check if there is a conflict in type annotations
        current_type = function.__annotations__.get(arg)
        if current_type and type and current_type != type:
            raise TypeError(
                "Argument {} in {} is both specified as {} "
                "and {}".format(
                    arg, function_to_str(function), current_type, type
                )
            )

        # if arg in function.__arguments_decorator_specs:
        #   raise ValueError(
        #     "@argument decorator was applied twice "
        #     "for the same argument {} on function {}".format(arg, function)
        #   )

        if positional and aliases:
            msg = ("Aliases are not yet supported for positional arguments"
                   "@ {}".format(arg))
            raise ValueError(msg)

        # reject positional=True if we are applied over a class
        if isclass(function) and positional:
            raise ValueError(
                "Cannot set positional arguments for super " "commands"
            )

        # We use __annotations__ to allow the usage of python 3 typing
        function.__annotations__.setdefault(arg, type)

        # pylint: disable=protected-access
        function.__arguments_decorator_specs[arg] = _ArgDecoratorSpec(
            arg=arg,
            description=description,
            name=name or transform_name(arg),
            aliases=aliases or [],
            positional=positional,
            choices=choices or [],
        )

        return function

    return decorator


def _init_attr(obj, attribute, default_value):
    if not hasattr(obj, attribute):
        setattr(obj, attribute, default_value)



"""StaticChunker module

This module contains a simple version a Chunker

Classes:
    StaticChunker: splits in a simple way the global inventory
                   in chunks

"""
from typing import Dict, List
import numpy as np
from suzieq.poller.controller.chunker.base_chunker \
    import Chunker
from suzieq.shared.exceptions import SqPollerConfError


class StaticChunker(Chunker):
    """StaticChunker splits the global inventory in chunks

    The StaticChunker supports policies for splitting:
    - sequential (default): splits the global inventory as is in chunks
    - namespace: splits the global inventory without splitting namespaces
    """

    def __init__(self, config_data: dict = None, validate: bool = True):
        super().__init__(config_data, validate)

        self.policies_list = ['sequential', 'namespace']
        self.policies_fn = {}

        for pol_name in self.policies_list:
            fun = getattr(self, f'split_{pol_name}', None)
            if not fun or not callable(fun):
                raise RuntimeError(f'Unknown {pol_name} policy')
            self.policies_fn[pol_name] = fun
        if config_data:
            policy = config_data \
                .get('policy', self.policies_list[0])
            if policy not in self.policies_list:
                raise SqPollerConfError(f'Unknown chunking policy {policy}')
            self.policy = policy
        else:
            self.policy = self.policies_list[0]

    @classmethod
    def get_data_model(cls):
        """This is only temporary. In future release I will add chunker
        validation via pydantic
        """
        raise NotImplementedError

    def chunk(self, glob_inv: dict, n_chunks: int, **kwargs) -> List[Dict]:

        policy = kwargs.pop('policy', self.policy)

        chunk_fun = self.policies_fn.get(policy, None)
        if not chunk_fun:
            raise SqPollerConfError(
                f'Unknown chunking function for policy {policy}')

        inv_chunks = [c for c in chunk_fun(glob_inv, n_chunks) if c]
        if len(inv_chunks) < n_chunks:
            if self.policy == 'sequential':
                raise SqPollerConfError(
                    'Not enough devices to split the inventory'
                    f'into {n_chunks} chunks'
                )
            if self.policy == 'namespace':
                raise SqPollerConfError(
                    'Not enough namespaces to split the inventory'
                    f'into {n_chunks} chunks'
                )
        return inv_chunks

    def split_sequential(self, glob_inv: dict, n_chunks: int) -> List[Dict]:
        """This function splits the global inventory following the
           same order of the global inventory. This function simply divides
           the global inventory as is "n_pollers" chunks without caring
           about, for example, splitting devices belonging to the same
           namespace

        Args:
            glob_inv (dict): the global inventory to split
            n_pollers (int): number of chunks to calculate

        Returns:
            List[Dict]: list of global_inventory chunks
        """
        chunk_len = int(len(glob_inv)/n_chunks)

        # leftovers: if the number of devices in the inventory is not divisible
        #            by "n_pollers" and the inventory is splitted in
        #            "n_pollers" chunks of size "chunk_len", the result would
        #            be that the last inventory will contain more devices than
        #            the others.
        #            The "leftover" variable is used to distribute these
        #            additional devices to all the inventories

        leftovers = len(glob_inv) % n_chunks
        inv_chunks = []
        chunk_start = 0
        for _ in range(n_chunks):
            if leftovers > 0:
                # leftovers must be divided between pollers
                inv_chunks.append(
                    dict(
                        list(glob_inv.items())
                        [chunk_start:chunk_start+chunk_len+1])
                )
                leftovers -= 1
                chunk_start += 1
            else:
                inv_chunks.append(
                    dict(
                        list(glob_inv.items())
                        [chunk_start:chunk_start+chunk_len])
                )
            chunk_start += chunk_len

        return inv_chunks

    def split_namespace(self, glob_inv: dict, n_pollers: int) -> List[Dict]:
        """split global inventory by namespace

        Args:
            glob_inv (dict): global inventory to split
            n_pollers (int): number of chunks to calculate

        Returns:
            List[Dict]: list of global_inventory chunks
        """

        def get_chunked_inventory(inventory_namespaces: dict, n_pollers: int):
            """
            This function splits the inventories trying to keep the same number
            of devices in each inventory chunk.

            namespaces_count contains the number of devices for each namespace
            chunk
            """
            namespaces_count = [0 for _ in range(n_pollers)]
            inventory_ns_chunks = [{} for _ in range(n_pollers)]

            for inv_ns_values in inventory_namespaces.values():
                min_index = np.argmin(namespaces_count)
                inventory_ns_chunks[min_index].update(inv_ns_values)
                namespaces_count[min_index] += len(inv_ns_values)

            return inventory_ns_chunks

        inventory_namespaces = {}

        for dev_name, device in glob_inv.items():
            namespace = device.get('namespace', None)
            if namespace is None:
                raise AttributeError(f"{dev_name} doesn't have namespace set")
            if not inventory_namespaces.get(namespace, None):
                inventory_namespaces[namespace] = {}
            inventory_namespaces[namespace][dev_name] = device

        return get_chunked_inventory(inventory_namespaces, n_pollers)



from suzieq.sqobjects.basicobj import SqObject


class OspfIfObj(SqObject):
    '''The object providing access to the ospf interfaces table

    Used only from tables object. Use ospf object otherwise.
    '''

    def __init__(self, **kwargs):
        super().__init__(table='ospfIf', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'ifname',
                                'vrf', 'area', 'state', 'query_str']



from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.lldp import LldpObj


@command("lldp", help="Act on LLDP data")
@argument("ifname", description="Interface name(s), space separated")
@argument("peerHostname", description="Peer hostname(s), space separated")
@argument("peerMacaddr", description="Peer mac address(es), space separated")
class LldpCmd(SqCommand):
    """LLDP protocol information"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = ' ',
            columns: str = "default",
            ifname: str = '',
            peerHostname: str = '',
            peerMacaddr: str = ''
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=LldpObj,
        )
        self.lvars = {
            'peerHostname': peerHostname.split(),
            'ifname': ifname.split(),
            'peerMacaddr': peerMacaddr.split(),
        }



# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument
import logging
from pathlib import Path
from typing import Dict, List, Optional
from urllib.parse import urlparse

from pydantic import BaseModel, Field, validator
from suzieq.poller.controller.source.base_source import Source, SourceModel
from suzieq.poller.controller.utils.inventory_utils import validate_hostname
from suzieq.shared.utils import PollerTransport

logger = logging.getLogger(__name__)

_DEFAULT_PORTS = {'http': 80, 'https': 443, 'ssh': 22}


class HostModel(BaseModel):
    """Model used to validate the hosts of a native inventory
    """
    username: str = Field(default=None)
    password: Optional[str] = Field(default=None)
    keyfile: Optional[str] = Field(default=None)
    devtype: Optional[str] = Field(default=None)
    port: str = Field(default=None)
    address: str = Field(default=None)
    transport: str = Field(default=None)
    url: str

    @validator('url')
    def validate_and_set(cls, url: str, values):
        """Validate the 'url' parameter and set the other parameters
        """
        words = url.split()
        decoded_url = urlparse(words[0])

        username = decoded_url.username
        password = (decoded_url.password or
                    # self.user_password or #I can't get this info here
                    None)
        address = decoded_url.hostname
        if not validate_hostname(address):
            raise ValueError(f'Invalid hostname or address {address}')
        transport = decoded_url.scheme or "https"
        try:
            PollerTransport[transport]
        except KeyError:
            raise ValueError(
                f"Transport '{transport}' not supported for host {address}")
        port = decoded_url.port or _DEFAULT_PORTS.get(transport)
        devtype = None
        keyfile = None

        try:
            for i in range(1, len(words[1:])+1):
                if words[i].startswith('keyfile'):
                    keyfile = words[i].split('=')[1]
                elif words[i].startswith('devtype'):
                    devtype = words[i].split('=')[1]
                elif words[i].startswith('username'):
                    username = words[i].split('=')[1]
                elif words[i].startswith('password'):
                    password = words[i].split('=')[1]
                else:
                    raise ValueError(
                        f'Unknown parameter: {words[i]} for {address}')
        except IndexError:
            if 'password' not in words[i]:
                raise ValueError(f"Missing '=' in key {words[i]}")
            raise ValueError("Invalid password spec., missing '='")

        if keyfile and not Path(keyfile).exists():
            raise ValueError(
                f"keyfile {keyfile} does not exist"
            )

        values['username'] = username
        values['password'] = password
        values['keyfile'] = keyfile
        values['devtype'] = devtype
        values['port'] = port
        values['address'] = address
        values['transport'] = transport

        return url


class NativeSourceModel(SourceModel):
    """Native source validation model
    """
    hosts: List[HostModel]

    @validator('hosts')
    def hosts_not_empty(cls, hosts: List):
        """checks if the hosts list is not empty"""
        if not hosts:
            raise ValueError('Empty hosts list')
        return hosts


class SqNativeFile(Source):
    """Source class used to load Suzieq native inventory files
    """

    def __init__(self, input_data, validate: bool = True) -> None:
        self._cur_inventory = {}
        super().__init__(input_data, validate)

    @classmethod
    def get_data_model(cls):
        return NativeSourceModel

    def _load(self, input_data):
        if not self._validate:
            new_hosts = []
            for h in input_data.get('hosts', []):
                new_hosts.append(HostModel(**h))
            input_data['hosts'] = new_hosts
        super()._load(input_data)

        self._cur_inventory = self._get_inventory()
        self.set_inventory(self._cur_inventory)

    def _get_inventory(self) -> Dict:
        """Extract the data from ansible inventory file

        Returns:
            Dict: inventory dictionary
        """
        inventory = {}

        for host in self._data.hosts:
            # I cannot use host.dict() due to field names changing
            entry = {
                'address': host.address,
                'username': host.username,
                'port': host.port,
                'password': host.password,
                'transport': host.transport,
                'devtype': host.devtype,
                'namespace': self._namespace,
                'ssh_keyfile': host.keyfile,
                'hostname': None,
            }
            inventory[f'{self._namespace}.{host.address}.{host.port}'] = entry

        if not inventory:
            logger.error('No hosts detected in provided inventory file')

        return inventory
